from ..models import State, Chamber, simple_numbered_districts

NY = State(
    name="New York",
    abbr="NY",
    capital="Albany",
    capital_tz="America/New_York",
    fips="36",
    unicameral=False,
    legislature_name="New York Legislature",
    legislature_organization_id="ocd-organization/700d3e60-3b1a-4cba-9aa4-06278aa7605d",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/4074536a-6e18-56f0-b004-3e85742b0c98",
    division_id="ocd-division/country:us/state:ny",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ny/government",
    url="http://public.leginfo.state.ny.us/",
    lower=Chamber(
        chamber_type="lower",
        name="Assembly",
        organization_id="ocd-organization/26bb6306-85f0-4d10-bff7-d1cd5bdc0865",
        num_seats=150,
        title="Assembly Member",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ny", "lower", 150
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/8291a233-623d-40e8-882d-21ec2d382c87",
        num_seats=63,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ny", "upper", 63
        ),
    ),
)



# Generated by Django 3.0.4 on 2020-04-06 11:32

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("data", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="RunPlan",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("success", models.BooleanField(default=True)),
                ("start_time", models.DateTimeField()),
                ("end_time", models.DateTimeField()),
                ("exception", models.TextField(blank=True, default="")),
                ("traceback", models.TextField(blank=True, default="")),
                (
                    "jurisdiction",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="runs",
                        to="data.Jurisdiction",
                    ),
                ),
            ],
            options={"db_table": "pupa_runplan"},
        ),
        migrations.CreateModel(
            name="SessionDataQualityReport",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("bills_missing_actions", models.PositiveIntegerField()),
                ("bills_missing_sponsors", models.PositiveIntegerField()),
                ("bills_missing_versions", models.PositiveIntegerField()),
                ("votes_missing_voters", models.PositiveIntegerField()),
                ("votes_missing_bill", models.PositiveIntegerField()),
                ("votes_missing_yes_count", models.PositiveIntegerField()),
                ("votes_missing_no_count", models.PositiveIntegerField()),
                ("votes_with_bad_counts", models.PositiveIntegerField()),
                (
                    "unmatched_sponsor_people",
                    django.contrib.postgres.fields.jsonb.JSONField(),
                ),
                (
                    "unmatched_sponsor_organizations",
                    django.contrib.postgres.fields.jsonb.JSONField(),
                ),
                ("unmatched_voters", django.contrib.postgres.fields.jsonb.JSONField()),
                (
                    "legislative_session",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="data.LegislativeSession",
                    ),
                ),
            ],
            options={"db_table": "pupa_sessiondataqualityreport"},
        ),
        migrations.CreateModel(
            name="ScrapeReport",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("scraper", models.CharField(max_length=300)),
                ("args", models.CharField(max_length=300)),
                ("start_time", models.DateTimeField()),
                ("end_time", models.DateTimeField()),
                (
                    "plan",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="scrapers",
                        to="reports.RunPlan",
                    ),
                ),
            ],
            options={"db_table": "pupa_scrapereport"},
        ),
        migrations.CreateModel(
            name="ScrapeObjects",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "object_type",
                    models.CharField(
                        choices=[
                            ("jurisdiction", "Jurisdiction"),
                            ("person", "Person"),
                            ("organization", "Organization"),
                            ("post", "Post"),
                            ("membership", "Membership"),
                            ("bill", "Bill"),
                            ("vote_event", "VoteEvent"),
                            ("event", "Event"),
                        ],
                        max_length=20,
                    ),
                ),
                ("count", models.PositiveIntegerField()),
                (
                    "report",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="scraped_objects",
                        to="reports.ScrapeReport",
                    ),
                ),
            ],
            options={"db_table": "pupa_scrapeobjects"},
        ),
        migrations.CreateModel(
            name="ImportObjects",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "object_type",
                    models.CharField(
                        choices=[
                            ("jurisdiction", "Jurisdiction"),
                            ("person", "Person"),
                            ("organization", "Organization"),
                            ("post", "Post"),
                            ("membership", "Membership"),
                            ("bill", "Bill"),
                            ("vote_event", "VoteEvent"),
                            ("event", "Event"),
                        ],
                        max_length=20,
                    ),
                ),
                ("insert_count", models.PositiveIntegerField()),
                ("update_count", models.PositiveIntegerField()),
                ("noop_count", models.PositiveIntegerField()),
                ("start_time", models.DateTimeField()),
                ("end_time", models.DateTimeField()),
                (
                    "report",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="imported_objects",
                        to="reports.RunPlan",
                    ),
                ),
            ],
            options={"db_table": "pupa_importobjects"},
        ),
        migrations.CreateModel(
            name="Identifier",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("identifier", models.CharField(max_length=300)),
                ("object_id", models.CharField(max_length=300)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "jurisdiction",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="pupa_ids",
                        to="data.Jurisdiction",
                    ),
                ),
            ],
            options={"db_table": "pupa_identifier"},
        ),
    ]



from .base import BaseImporter
from ._types import _JsonDict
from ..utils import get_pseudo_id, _make_pseudo_id
from ..data.models import (
    Event,
    EventLocation,
    EventSource,
    EventDocument,
    EventDocumentLink,
    EventLink,
    EventParticipant,
    EventMedia,
    EventMediaLink,
    EventAgendaItem,
    EventRelatedEntity,
    EventAgendaMedia,
    EventAgendaMediaLink,
)
from .organizations import OrganizationImporter
from .vote_events import VoteEventImporter
from .bills import BillImporter


class EventImporter(BaseImporter):
    _type = "event"
    model_class = Event
    related_models = {
        "sources": (EventSource, "event_id", {}),
        "documents": (
            EventDocument,
            "event_id",
            {"links": (EventDocumentLink, "document_id", {})},
        ),
        "links": (EventLink, "event_id", {}),
        "participants": (EventParticipant, "event_id", {}),
        "media": (EventMedia, "event_id", {"links": (EventMediaLink, "media_id", {})}),
        "agenda": (
            EventAgendaItem,
            "event_id",
            {
                "related_entities": (EventRelatedEntity, "agenda_item_id", {}),
                "media": (
                    EventAgendaMedia,
                    "agenda_item_id",
                    {"links": (EventAgendaMediaLink, "media_id", {})},
                ),
            },
        ),
    }
    preserve_order = {"agenda"}

    def __init__(
        self,
        jurisdiction_id: str,
        org_importer: OrganizationImporter,
        bill_importer: BillImporter,
        vote_event_importer: VoteEventImporter,
    ):
        super(EventImporter, self).__init__(jurisdiction_id)
        self.org_importer = org_importer
        self.bill_importer = bill_importer
        self.vote_event_importer = vote_event_importer

    def get_object(self, event: _JsonDict) -> Event:
        if event.get("dedupe_key"):
            spec = {
                "dedupe_key": event.get("dedupe_key"),
                "jurisdiction_id": self.jurisdiction_id,
            }
        else:
            spec = {
                "name": event["name"],
                "description": event["description"],
                "start_date": event["start_date"],
                "end_date": event["end_date"],
                "jurisdiction_id": self.jurisdiction_id,
            }
        return self.model_class.objects.get(**spec)

    def get_location(self, location_data: _JsonDict) -> EventLocation:
        obj, created = EventLocation.objects.get_or_create(
            name=location_data["name"],
            url=location_data.get("url", ""),
            jurisdiction_id=self.jurisdiction_id,
        )
        # TODO: geocode here?
        return obj

    def prepare_for_db(self, data: _JsonDict) -> _JsonDict:
        data["jurisdiction_id"] = self.jurisdiction_id
        data["location"] = self.get_location(data["location"])

        data["start_date"] = data["start_date"]
        data["end_date"] = data.get("end_date", "")

        for participant in data["participants"]:
            if "person_id" in participant:
                participant["person_id"] = self.resolve_person(participant["person_id"])
            elif "organization_id" in participant:
                participant["organization_id"] = self.org_importer.resolve_json_id(
                    participant["organization_id"], allow_no_match=True
                )

        for item in data["agenda"]:
            for entity in item["related_entities"]:
                if "person_id" in entity:
                    entity["person_id"] = self.resolve_person(entity["person_id"])
                elif "organization_id" in entity:
                    entity["organization_id"] = self.org_importer.resolve_json_id(
                        entity["organization_id"], allow_no_match=True
                    )
                elif "bill_id" in entity:
                    # unpack and repack bill psuedo id in case filters alter it
                    bill = get_pseudo_id(entity["bill_id"])
                    self.bill_importer.apply_transformers(bill)
                    bill = _make_pseudo_id(**bill)
                    entity["bill_id"] = self.bill_importer.resolve_json_id(
                        bill, allow_no_match=True
                    )
                elif "vote_event_id" in entity:
                    entity["vote_event_id"] = self.vote_event_importer.resolve_json_id(
                        entity["vote_event_id"], allow_no_match=True
                    )

        return data



from ..models import State, Chamber, simple_numbered_districts

MS = State(
    name="Mississippi",
    abbr="MS",
    capital="Jackson",
    capital_tz="America/Chicago",
    fips="28",
    unicameral=False,
    legislature_name="Mississippi Legislature",
    legislature_organization_id="ocd-organization/f2b7fee5-8da1-431d-8ea1-19b7709194cc",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/2d23c3ec-897d-4ef8-85a5-0acb24328d5f",
    division_id="ocd-division/country:us/state:ms",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ms/government",
    url="http://www.legislature.ms.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/47c2e037-ffe3-46c3-9399-8feb1f7d4988",
        num_seats=122,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ms", "lower", 122
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/7e3c07c9-91ff-4445-8cef-57f85ab6ae9f",
        num_seats=52,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ms", "upper", 52
        ),
    ),
)



from .common import extras, fuzzy_date_blank

schema = {
    "properties": {
        "label": {"type": "string"},
        "role": {"type": "string"},
        "person_id": {"type": ["string", "null"]},
        "person_name": {"type": ["string"], "minLength": 1},
        "organization_id": {"type": "string", "minLength": 1},
        "post_id": {"type": ["string", "null"]},
        "start_date": fuzzy_date_blank,
        "end_date": fuzzy_date_blank,
        "extras": extras,
        # division & jurisdiction are additions to popolo
        "division_id": {"type": ["string", "null"]},
        "jurisdiction_id": {"type": "string", "minLength": 1},
    },
    "type": "object",
}



# Generated by Django 3.0.5 on 2021-02-10 18:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0029_auto_20210203_1946"),
    ]

    operations = [
        migrations.AlterField(
            model_name="relatedbill",
            name="relation_type",
            field=models.CharField(
                choices=[
                    ("companion", "Companion"),
                    ("prior-session", "Prior Session"),
                    ("replaced-by", "Replaced By"),
                    ("replaces", "Replaces"),
                    ("related", "Related"),
                ],
                max_length=100,
            ),
        ),
    ]



from .common import links, contact_details, extras, fuzzy_date_blank

schema = {
    "properties": {
        "label": {"type": "string", "minLength": 1},
        "role": {"type": "string"},
        "maximum_memberships": {"type": "number"},
        "organization_id": {"type": "string", "minLength": 1},
        "division_id": {"type": ["null", "string"], "minLength": 1},
        "start_date": fuzzy_date_blank,
        "end_date": fuzzy_date_blank,
        "contact_details": contact_details,
        "links": links,
        "extras": extras,
    },
    "type": "object",
}



import pytest
from openstates.scrape import VoteEvent, Bill, Organization
from openstates.utils import get_pseudo_id


def toy_vote_event():
    ve = VoteEvent(
        legislative_session="2009",
        motion_text="passage of the bill",
        start_date="2009-01-07",
        result="pass",
        classification="passage",
    )
    ve.add_source("http://uri.example.com/", note="foo")
    return ve


def test_simple_vote_event():
    ve = toy_vote_event()
    ve.set_count("yes", 2)
    ve.yes("James")
    ve.no("Paul")
    ve.vote("abstain", "Thom")

    assert len(ve.votes) == 3
    assert len(ve.counts) == 1
    assert get_pseudo_id(ve.organization) == {"classification": "legislature"}
    assert get_pseudo_id(ve.votes[0]["voter_id"]) == {"name": "James"}
    assert get_pseudo_id(ve.votes[1]["voter_id"]) == {"name": "Paul"}
    assert get_pseudo_id(ve.votes[2]["voter_id"]) == {"name": "Thom"}
    assert ve.bill is None

    ve.validate()
    assert "we get here"


def test_vote_event_org_obj():
    o = Organization("something", classification="committee")
    ve = VoteEvent(
        legislative_session="2009",
        motion_text="passage of the bill",
        start_date="2009-01-07",
        result="pass",
        classification="passage",
        organization=o,
    )
    assert ve.organization == o._id


def test_vote_event_org_dict():
    odict = {"name": "Random Committee", "classification": "committee"}
    ve = VoteEvent(
        legislative_session="2009",
        motion_text="passage of the bill",
        start_date="2009-01-07",
        result="pass",
        classification="passage",
        organization=odict,
    )
    assert get_pseudo_id(ve.organization) == odict


def test_vote_event_org_chamber():
    ve = VoteEvent(
        legislative_session="2009",
        motion_text="passage of the bill",
        start_date="2009-01-07",
        result="pass",
        classification="passage",
        chamber="upper",
    )
    assert get_pseudo_id(ve.organization) == {"classification": "upper"}


def test_org_and_chamber_conflict():
    with pytest.raises(ValueError):
        VoteEvent(
            legislative_session="2009",
            motion_text="passage of the bill",
            start_date="2009-01-07",
            result="pass",
            classification="passage",
            organization="test",
            chamber="lower",
        )


def test_set_count():
    ve = toy_vote_event()
    ve.set_count("yes", 2)
    ve.set_count("no", 100)
    ve.set_count("yes", 0)
    assert ve.counts == [{"option": "yes", "value": 0}, {"option": "no", "value": 100}]


def test_set_bill_obj():
    ve = toy_vote_event()
    b = Bill("HB 1", legislative_session="2009", title="fake bill")
    ve.set_bill(b)
    assert ve.bill == b._id


def test_set_bill_obj_no_extra_args():
    ve = toy_vote_event()
    b = Bill("HB 1", legislative_session="2009", title="fake bill")
    with pytest.raises(ValueError):
        ve.set_bill(b, chamber="lower")


def test_set_bill_pseudo_id():
    ve = toy_vote_event()
    ve.set_bill("HB 1", chamber="lower")
    assert get_pseudo_id(ve.bill) == {
        "identifier": "HB 1",
        "from_organization__classification": "lower",
        "legislative_session__identifier": "2009",
    }


def test_str():
    ve = toy_vote_event()
    s = str(ve)
    assert ve.legislative_session in s
    assert ve.motion_text in s



from .common import (
    links,
    contact_details,
    identifiers,
    other_names,
    sources,
    extras,
    fuzzy_date_blank,
)

schema = {
    "properties": {
        "name": {"type": "string", "minLength": 1},
        "other_names": other_names,
        "identifiers": identifiers,
        "family_name": {"type": "string"},
        "given_name": {"type": "string"},
        "gender": {"type": "string"},
        "birth_date": fuzzy_date_blank,
        "death_date": fuzzy_date_blank,
        "image": {"format": "uri-blank", "type": "string"},
        "biography": {"type": "string"},
        "contact_details": contact_details,
        "links": links,
        "sources": sources,
        "extras": extras,
    },
    "type": "object",
}



"""
    Schema for bill objects.
"""
import copy
from .common import sources, extras, fuzzy_date_blank, fuzzy_datetime
from ...data import common

versions_or_documents = {
    "items": {
        "properties": {
            "note": {"type": "string", "minLength": 1},
            "date": fuzzy_date_blank,
            "classification": {"type": "string"},
            "links": {
                "items": {
                    "properties": {
                        "media_type": {"type": "string"},
                        "url": {"type": "string", "format": "uri"},
                    },
                    "type": "object",
                },
                "type": "array",
            },
        },
        "type": "object",
    },
    "type": "array",
}
versions = copy.deepcopy(versions_or_documents)
versions["items"]["properties"]["classification"][
    "enum"
] = common.BILL_VERSION_CLASSIFICATIONS
documents = copy.deepcopy(versions_or_documents)
documents["items"]["properties"]["classification"][
    "enum"
] = common.BILL_DOCUMENT_CLASSIFICATIONS

schema = {
    "type": "object",
    "properties": {
        "legislative_session": {"type": "string", "minLength": 1},
        "identifier": {"type": "string", "minLength": 1},
        "title": {"type": "string", "minLength": 1},
        "from_organization": {"type": ["string", "null"]},
        "classification": {
            "items": {"type": "string", "enum": common.BILL_CLASSIFICATIONS},
            "type": "array",
        },
        "subject": {"items": {"type": "string", "minLength": 1}, "type": "array"},
        "abstracts": {
            "items": {
                "properties": {
                    "abstract": {"type": "string", "minLength": 1},
                    "note": {"type": "string"},
                    "date": {"type": "string"},
                },
                "type": "object",
            },
            "type": "array",
        },
        "other_titles": {
            "items": {
                "properties": {
                    "title": {"type": "string", "minLength": 1},
                    "note": {"type": "string"},
                },
                "type": "object",
            },
            "type": "array",
        },
        "other_identifiers": {
            "items": {
                "properties": {
                    "identifier": {"type": "string", "minLength": 1},
                    "note": {"type": "string"},
                    "scheme": {"type": "string"},
                },
                "type": "object",
            },
            "type": "array",
        },
        "actions": {
            "items": {
                "properties": {
                    "organization": {"type": ["string", "null"]},
                    "date": fuzzy_datetime,
                    "description": {"type": "string", "minLength": 1},
                    "classification": {
                        "items": {
                            "type": "string",
                            "enum": common.BILL_ACTION_CLASSIFICATIONS,
                        },
                        "type": "array",
                    },
                    "related_entities": {
                        "items": {
                            "properties": {
                                "name": {"type": "string", "minLength": 1},
                                "entity_type": {
                                    "enum": ["organization", "person", ""],
                                    "type": "string",
                                },
                                "person_id": {"type": ["string", "null"]},
                                "organization_id": {"type": ["string", "null"]},
                            },
                            "type": "object",
                        },
                        "type": "array",
                    },
                },
                "type": "object",
            },
            "type": "array",
        },
        "sponsorships": {
            "items": {
                "properties": {
                    "primary": {"type": "boolean"},
                    "classification": {"type": "string", "minLength": 1},
                    "name": {"type": "string", "minLength": 1},
                    "entity_type": {
                        "enum": ["organization", "person", ""],
                        "type": "string",
                    },
                    "person_id": {"type": ["string", "null"]},
                    "organization_id": {"type": ["string", "null"]},
                },
                "type": "object",
            },
            "type": "array",
        },
        "related_bills": {
            "items": {
                "properties": {
                    "identifier": {"type": "string", "minLength": 1},
                    "legislative_session": {"type": "string", "minLength": 1},
                    "relation_type": {
                        "enum": common.BILL_RELATION_TYPES,
                        "type": "string",
                    },
                },
                "type": "object",
            },
            "type": "array",
        },
        "versions": versions,
        "documents": documents,
        "sources": sources,
        "extras": extras,
    },
}



import os
import copy
import glob
import json
import logging
import typing
from django.db.models import Q, Model
from django.db.models.signals import post_save
from .. import settings
from ..data.models import LegislativeSession, Person
from ..exceptions import DuplicateItemError, UnresolvedIdError, DataImportError
from ..utils import get_pseudo_id, utcnow
from ._types import _ID, _JsonDict, _RelatedModels, _TransformerMapping

_PersonCacheKey = typing.Tuple[str, typing.Optional[str], typing.Optional[str]]


def omnihash(obj: typing.Any) -> int:
    """ recursively hash unhashable objects """
    if isinstance(obj, set):
        return hash(frozenset(omnihash(e) for e in obj))
    elif isinstance(obj, (tuple, list)):
        return hash(tuple(omnihash(e) for e in obj))
    elif isinstance(obj, dict):
        return hash(frozenset((k, omnihash(v)) for k, v in obj.items()))
    else:
        return hash(obj)


def _match(
    dbitem: Model,
    jsonitem: _JsonDict,
    keys: typing.Iterable[str],
    subfield_dict: typing.Dict[str, typing.Any],
) -> bool:
    # check if all keys (excluding subfields) match
    for k in keys:
        if k not in subfield_dict and getattr(dbitem, k) != jsonitem.get(k, None):
            return False

    # all fields match so far, possibly equal, just check subfields now
    for k in subfield_dict:
        jsonsubitems = jsonitem[k]
        dbsubitems = list(getattr(dbitem, k).all())
        if items_differ(jsonsubitems, dbsubitems, subfield_dict[k][2]):
            return False

    # if we got here, item values match
    return True


def items_differ(
    jsonitems: typing.List[_JsonDict],
    dbitems: typing.List[Model],
    subfield_dict: _JsonDict,
) -> bool:
    """ check whether or not jsonitems and dbitems differ """

    # short circuit common cases
    if len(jsonitems) == len(dbitems) == 0:
        # both are empty
        return False
    elif len(jsonitems) != len(dbitems):
        # if lengths differ, they're definitely different
        return True

    original_jsonitems = jsonitems
    jsonitems = copy.deepcopy(jsonitems)
    keys = jsonitems[0].keys()

    # go over dbitems looking for matches
    for dbitem in dbitems:
        order = getattr(dbitem, "order", None)

        match = None

        # if we have an order, we can just check one item
        if order is not None:
            # use original so that pop calls don't affect ordering
            if not _match(dbitem, original_jsonitems[order], keys, subfield_dict):
                # short circuit if there isn't a match in the right spot
                return True

        # need to get position of match to remove
        for i, jsonitem in enumerate(jsonitems):
            if _match(dbitem, jsonitem, keys, subfield_dict):
                match = i

        if match is not None:
            # item exists in both, remove from jsonitems
            jsonitems.pop(match)
        else:
            # exists in db but not json
            return True

    # if we get here, jsonitems has to be empty because we asserted that the length was
    # the same and we found a match for each thing in dbitems, here's a safety check just in case
    if jsonitems:  # pragma: no cover
        return True

    return False


class BaseImporter:
    """BaseImporter

    Override:
        get_object(data)
        limit_spec(spec)                [optional, required if pseudo_ids are used]
        prepare_for_db(data)            [optional]
        postimport()                    [optional]
        update_computed_fields(obj)     [optional]
    """

    _type: str
    model_class: Model = None
    related_models: _RelatedModels = {}
    preserve_order: typing.Set[str] = set()
    merge_related: typing.Dict[str, typing.List[str]] = {}
    cached_transformers: _TransformerMapping = {}

    def __init__(self, jurisdiction_id: str) -> None:
        self.jurisdiction_id = jurisdiction_id
        self.json_to_db_id: typing.Dict[str, _ID] = {}
        self.duplicates: typing.Dict[str, str] = {}
        self.pseudo_id_cache: typing.Dict[str, typing.Optional[_ID]] = {}
        self.person_cache: typing.Dict[_PersonCacheKey, typing.Optional[str]] = {}
        self.session_cache: typing.Dict[str, LegislativeSession] = {}
        self.logger = logging.getLogger("openstates")
        self.info = self.logger.info
        self.debug = self.logger.debug
        self.warning = self.logger.warning
        self.error = self.logger.error
        self.critical = self.logger.critical

        # load transformers from appropriate setting
        if settings.IMPORT_TRANSFORMERS.get(self._type):
            self.cached_transformers = settings.IMPORT_TRANSFORMERS[self._type]

    def get_session(self, identifier: str) -> LegislativeSession:
        if identifier not in self.session_cache:
            self.session_cache[identifier] = LegislativeSession.objects.get(
                identifier=identifier, jurisdiction_id=self.jurisdiction_id
            )
        return self.session_cache[identifier]

    def limit_spec(self, spec: _JsonDict) -> _JsonDict:
        raise NotImplementedError()

    def get_object(self, object: _JsonDict) -> Model:
        raise NotImplementedError()

    # no-ops to be overriden
    def prepare_for_db(self, data: _JsonDict) -> _JsonDict:
        return data

    def postimport(self) -> None:
        pass

    def update_computed_fields(self, obj: Model) -> None:
        pass

    def resolve_json_id(
        self, json_id: str, allow_no_match: bool = False
    ) -> typing.Optional[_ID]:
        """
        Given an id found in scraped JSON, return a DB id for the object.

        params:
            json_id:        id from json
            allow_no_match: just return None if id can't be resolved

        returns:
            database id

        raises:
            ValueError if id couldn't be resolved
        """
        if not json_id:
            return None

        if json_id.startswith("~"):
            # keep caches of all the pseudo-ids to avoid doing 1000s of lookups during import
            if json_id not in self.pseudo_id_cache:
                spec = get_pseudo_id(json_id)
                spec = self.limit_spec(spec)

                if isinstance(spec, Q):
                    objects = self.model_class.objects.filter(spec)
                else:
                    objects = self.model_class.objects.filter(**spec)
                ids = {each.id for each in objects}
                if len(ids) == 1:
                    self.pseudo_id_cache[json_id] = ids.pop()
                    errmsg = None
                elif not ids:
                    errmsg = "cannot resolve pseudo id to {}: {}".format(
                        self.model_class.__name__, json_id
                    )
                else:
                    errmsg = "multiple objects returned for {} pseudo id {}: {}".format(
                        self.model_class.__name__, json_id, ids
                    )

                # either raise or log error
                if errmsg:
                    if not allow_no_match:
                        raise UnresolvedIdError(errmsg)
                    else:
                        self.error(errmsg)
                        self.pseudo_id_cache[json_id] = None

            # return the cached object
            return self.pseudo_id_cache[json_id]

        # get the id that the duplicate points to, or use self
        json_id = self.duplicates.get(json_id, json_id)

        try:
            return self.json_to_db_id[json_id]
        except KeyError:
            raise UnresolvedIdError("cannot resolve id: {}".format(json_id))

    def import_directory(self, datadir: str) -> typing.Dict[str, typing.Dict]:
        """ import a JSON directory into the database """

        def json_stream() -> typing.Iterator[_JsonDict]:
            # load all json, mapped by json_id
            for fname in glob.glob(os.path.join(datadir, self._type + "_*.json")):
                with open(fname) as f:
                    yield json.load(f)

        return self.import_data(json_stream())

    def _prepare_imports(
        self, dicts: typing.Iterable[_JsonDict]
    ) -> typing.Iterator[typing.Tuple[str, _JsonDict]]:
        """filters the import stream to remove duplicates

        also serves as a good place to override if anything special has to be done to the
        order of the import stream (see OrganizationImporter)
        """
        # hash(json): id
        seen_hashes = {}

        for data in dicts:
            json_id = data.pop("_id")
            if self._type == "vote_event":
                data.pop("bill_identifier", None)

            # map duplicates (using omnihash to tell if json dicts are identical-ish)
            objhash = omnihash(data)
            if objhash not in seen_hashes:
                seen_hashes[objhash] = json_id
                yield json_id, data
            else:
                self.duplicates[json_id] = seen_hashes[objhash]

    def import_data(
        self, data_items: typing.Iterable[_JsonDict]
    ) -> typing.Dict[str, typing.Dict]:
        """ import a bunch of dicts together """
        # keep counts of all actions
        record = {
            "insert": 0,
            "update": 0,
            "noop": 0,
            "start": utcnow(),
            "records": {"insert": [], "update": [], "noop": []},
        }

        for json_id, data in self._prepare_imports(data_items):
            obj_id, what = self.import_item(data)
            self.json_to_db_id[json_id] = obj_id
            record["records"][what].append(obj_id)
            record[what] += 1

        # all objects are loaded, a perfect time to do inter-object resolution and other tasks
        self.postimport()

        record["end"] = utcnow()

        return {self._type: record}

    def import_item(self, data: _JsonDict) -> typing.Tuple[_ID, str]:
        """ function used by import_data """
        what = "noop"

        # remove the JSON _id (may still be there if called directly)
        data.pop("_id", None)
        if self._type == "vote_event":
            data.pop("bill_identifier", None)

        # add fields/etc.
        data = self.apply_transformers(data)
        data = self.prepare_for_db(data)

        try:
            obj = self.get_object(data)
        except self.model_class.DoesNotExist:
            obj = None

        # pull related fields off
        related = {}
        for field in self.related_models:
            related[field] = data.pop(field)

        # obj existed, check if we need to do an update
        if obj:
            if obj.id in self.json_to_db_id.values():
                raise DuplicateItemError(data, obj, related.get("sources", []))
            # check base object for changes
            for key, value in data.items():
                if getattr(obj, key) != value:
                    setattr(obj, key, value)
                    what = "update"

            updated = self._update_related(obj, related, self.related_models)
            if updated:
                what = "update"

            if what == "update":
                # make sure to do this after create related
                self.update_computed_fields(obj)
                obj.save()

        # need to create the data
        else:
            what = "insert"
            try:
                obj = self.model_class(**data)
                obj.save()
            except Exception as e:
                raise DataImportError(
                    "{} while importing {} as {}".format(e, data, self.model_class)
                )
            self._create_related(obj, related, self.related_models)

            # make sure to do this after create related
            self.update_computed_fields(obj)

            # Fire post-save signal after related objects are created to allow
            # for handlers make use of related objects
            post_save.send(sender=self.model_class, instance=obj, created=True)

        return obj.id, what

    def _update_related(
        self,
        obj: Model,
        related: typing.Dict[str, typing.List[Model]],
        subfield_dict: _JsonDict,
    ) -> bool:
        """
        update DB objects related to a base object
            obj:            a base object to create related
            related:        dict mapping field names to lists of related objects
            subfield_list:  where to get the next layer of subfields
        """
        # keep track of whether or not anything was updated
        updated = False

        # for each related field - check if there are differences
        for field, items in related.items():
            # get items from database
            dbitems = list(getattr(obj, field).all())
            dbitems_count = len(dbitems)

            # default to doing nothing
            do_delete = do_update = False

            if items and dbitems_count:  # we have items, so does db, check for conflict
                do_delete = do_update = items_differ(
                    items, dbitems, subfield_dict[field][2]
                )
            elif items and not dbitems_count:  # we have items, db doesn't, just update
                do_update = True
            elif not items and dbitems_count:  # db has items, we don't, just delete
                do_delete = True
            # otherwise: no items or dbitems, so nothing is done

            # don't delete if field is in merge_related
            if field in self.merge_related:
                new_items = []
                # build a list of keyfields to existing database objects
                keylist = self.merge_related[field]
                keyed_dbitems = {
                    tuple(getattr(item, k) for k in keylist): item for item in dbitems
                }

                # go through 'new' items
                #   if item with the same keyfields exists:
                #       update the database item w/ the new item's properties
                #   else:
                #       add it to new_items
                for item in items:
                    key = tuple(item.get(k) for k in keylist)
                    dbitem = keyed_dbitems.get(key)
                    if not dbitem:
                        new_items.append(item)
                    else:
                        # update dbitem
                        for fname, val in item.items():
                            setattr(dbitem, fname, val)
                        dbitem.save()

                # import anything that made it to new_items in the usual fashion
                self._create_related(obj, {field: new_items}, subfield_dict)
            else:
                # default logic is to just wipe and recreate subobjects
                if do_delete:
                    updated = True
                    getattr(obj, field).all().delete()
                if do_update:
                    updated = True
                    self._create_related(obj, {field: items}, subfield_dict)

        return updated

    def _create_related(
        self,
        obj: Model,
        related: typing.Dict[str, typing.List[Model]],
        subfield_dict: _JsonDict,
    ) -> None:
        """
        create DB objects related to a base object
            obj:            a base object to create related
            related:        dict mapping field names to lists of related objects
            subfield_list:  where to get the next layer of subfields
        """
        for field, items in related.items():
            subobjects = []
            all_subrelated = []
            Subtype, reverse_id_field, subsubdict = subfield_dict[field]
            for order, item in enumerate(items):
                # pull off 'subrelated' (things that are related to this obj)
                subrelated = {}
                for subfield in subsubdict:
                    subrelated[subfield] = item.pop(subfield)

                if field in self.preserve_order:
                    item["order"] = order

                item[reverse_id_field] = obj.id

                try:
                    subobjects.append(Subtype(**item))
                    all_subrelated.append(subrelated)
                except Exception as e:
                    raise DataImportError(
                        "{} while importing {} as {}".format(e, item, Subtype)
                    )

            # add all subobjects at once (really great for actions & votes)
            try:
                Subtype.objects.bulk_create(subobjects)
            except Exception as e:
                raise DataImportError(
                    "{} while importing {} as {}".format(e, subobjects, Subtype)
                )

            # after import the subobjects, import their subsubobjects
            for subobj, subrel in zip(subobjects, all_subrelated):
                self._create_related(subobj, subrel, subsubdict)

    def apply_transformers(
        self, data: _JsonDict, transformers: typing.Optional[_TransformerMapping] = None
    ) -> _JsonDict:
        if transformers is None:
            transformers = self.cached_transformers

        for key, key_transformers in transformers.items():
            if key not in data:
                continue
            if isinstance(key_transformers, list):
                for transformer in key_transformers:
                    data[key] = transformer(data[key])
            elif isinstance(key_transformers, dict):
                self.apply_transformers(data[key], key_transformers)
            else:
                data[key] = key_transformers(data[key])

        return data

    def get_seen_sessions(self) -> typing.List[int]:
        return [s.id for s in self.session_cache.values()]

    def resolve_person(
        self,
        psuedo_person_id: str,
        start_date: typing.Optional[str] = None,
        end_date: typing.Optional[str] = None,
        org_classification: typing.Optional[str] = None,
    ) -> str:
        cache_key = (psuedo_person_id, start_date, end_date)
        if cache_key in self.person_cache:
            return self.person_cache[cache_key]

        # turn spec into DB query
        spec = get_pseudo_id(psuedo_person_id)
        if list(spec.keys()) == ["name"]:
            # if we're just resolving on name, include other names and family name
            name = spec["name"]
            spec = (
                Q(name__iexact=name)
                | Q(other_names__name__iexact=name)
                | Q(family_name__iexact=name)
            )
        else:
            spec = Q(**spec)

        spec &= Q(
            memberships__organization__jurisdiction_id=self.jurisdiction_id,
        )

        if org_classification:
            spec &= Q(memberships__organization__classification=org_classification)
        else:
            spec &= Q(
                memberships__organization__classification__in=(
                    "upper",
                    "lower",
                    "legislature",
                )
            )

        # we don't know what dates we have available to us... it can be any configuration
        # of start/end dates or they can all be null.  Instead of requiring a strict overlap
        # we use them to exclude people that are definitely not serving in the session:
        # if we know when the session started, ensure they are either in office still or
        #   that their end date was after the start of the session
        if start_date:
            spec &= Q(memberships__end_date="") | Q(
                memberships__end_date__gt=start_date
            )
        # if we know when the session ended, ensure that they didn't start after it ended
        if end_date:
            spec &= Q(memberships__start_date="") | Q(
                memberships__start_date__lt=end_date
            )

        ids = set(Person.objects.filter(spec).values_list("id", flat=True))
        if len(ids) == 1:
            self.person_cache[cache_key] = ids.pop()
            errmsg = None
        elif not ids:
            errmsg = f"no people returned for {spec}"
        else:
            errmsg = f"multiple people returned for {spec}"

        # either raise or log error
        if errmsg:
            self.error(errmsg)
            self.person_cache[cache_key] = None

        # return the newly-cached object
        return self.person_cache[cache_key]



import pytest
from openstates.scrape import Event as ScrapeEvent
from openstates.importers import (
    EventImporter,
    OrganizationImporter,
    BillImporter,
    VoteEventImporter,
)
from openstates.data.models import (
    Person,
    Membership,
    Organization,
    Jurisdiction,
    Division,
    VoteEvent,
    Bill,
    Event,
)


def create_jurisdiction():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    j = Jurisdiction.objects.create(id="jid", division_id="ocd-division/country:us")
    return j


def create_other_jurisdiction():
    Division.objects.create(id="ocd-division/country:ca", name="USA")
    j = Jurisdiction.objects.create(id="ojid", division_id="ocd-division/country:ca")
    return j


def ge():
    event = ScrapeEvent(
        name="America's Birthday",
        start_date="2014-07-04T05:00Z",
        location_name="America",
        all_day=True,
    )
    return event


oi = OrganizationImporter("jid")
bi = BillImporter("jid")
vei = VoteEventImporter("jid", bi)


@pytest.mark.django_db
def test_related_people_event():
    create_jurisdiction()
    george = Person.objects.create(id="gw", name="George Washington")
    john = Person.objects.create(id="jqp", name="John Q. Public")
    o = Organization.objects.create(
        name="Merica", jurisdiction_id="jid", classification="legislature"
    )

    Membership.objects.create(person=george, organization=o)
    Membership.objects.create(person=john, organization=o)

    event1 = ge()
    event2 = ge()

    for event in [event1, event2]:
        item = event.add_agenda_item("Cookies will be served")
        item.add_person(person="John Q. Public")
        event.add_person("George Washington")

    result = EventImporter("jid", oi, bi, vei).import_data([event1.as_dict()])
    assert result["event"]["insert"] == 1

    result = EventImporter("jid", oi, bi, vei).import_data([event2.as_dict()])
    assert result["event"]["noop"] == 1

    assert (
        Event.objects.get(name="America's Birthday").participants.first().person_id
        == "gw"
    )

    assert (
        Event.objects.get(name="America's Birthday")
        .agenda.first()
        .related_entities.first()
        .person_id
        == "jqp"
    )


@pytest.mark.django_db
def test_related_vote_event():
    j = create_jurisdiction()
    session = j.legislative_sessions.create(name="1900", identifier="1900")
    org = Organization.objects.create(id="org-id", name="House", classification="lower")
    bill = Bill.objects.create(
        id="bill-1", identifier="HB 1", legislative_session=session
    )
    VoteEvent.objects.create(
        id="vote-1",
        identifier="Roll no. 12",
        bill=bill,
        legislative_session=session,
        organization=org,
    )

    event1 = ge()
    event2 = ge()

    for event in [event1, event2]:
        item = event.add_agenda_item("Cookies will be served")
        item.add_vote_event(vote_event="Roll no. 12")

    result = EventImporter("jid", oi, bi, vei).import_data([event1.as_dict()])
    assert result["event"]["insert"] == 1

    result = EventImporter("jid", oi, bi, vei).import_data([event2.as_dict()])
    assert result["event"]["noop"] == 1

    assert (
        Event.objects.get(name="America's Birthday")
        .agenda.first()
        .related_entities.first()
        .vote_event_id
        == "vote-1"
    )


@pytest.mark.django_db
def test_related_bill_event():
    j = create_jurisdiction()
    session = j.legislative_sessions.create(name="1900", identifier="1900")
    Organization.objects.create(id="org-id", name="House", classification="lower")
    Bill.objects.create(id="bill-1", identifier="HB 101", legislative_session=session)
    event1 = ge()
    event2 = ge()

    for event in [event1, event2]:
        item = event.add_agenda_item("Cookies will be served")
        item.add_bill(bill="HB 101")

    result = EventImporter("jid", oi, bi, vei).import_data([event1.as_dict()])
    assert result["event"]["insert"] == 1

    result = EventImporter("jid", oi, bi, vei).import_data([event2.as_dict()])
    assert result["event"]["noop"] == 1

    assert (
        Event.objects.get(name="America's Birthday")
        .agenda.first()
        .related_entities.first()
        .bill_id
        == "bill-1"
    )


@pytest.mark.django_db
def test_related_committee_event():
    j = create_jurisdiction()
    j.legislative_sessions.create(name="1900", identifier="1900")
    org = Organization.objects.create(
        id="org-id", name="House", classification="lower", jurisdiction=j
    )
    Organization.objects.create(
        id="fiscal",
        name="Fiscal Committee",
        classification="committee",
        parent=org,
        jurisdiction=j,
    )

    event1 = ge()
    event2 = ge()

    for event in [event1, event2]:
        item = event.add_agenda_item("Cookies will be served")
        item.add_committee(committee="Fiscal Committee")

    result = EventImporter("jid", oi, bi, vei).import_data([event1.as_dict()])
    assert result["event"]["insert"] == 1

    result = EventImporter("jid", oi, bi, vei).import_data([event2.as_dict()])
    assert result["event"]["noop"] == 1

    assert (
        Event.objects.get(name="America's Birthday")
        .agenda.first()
        .related_entities.first()
        .organization_id
        == "fiscal"
    )


@pytest.mark.django_db
def test_media_event():
    create_jurisdiction()
    event1 = ge()
    event2 = ge()

    for event in [event1, event2]:
        item = event.add_agenda_item("Cookies will be served")
        item.add_media_link(
            note="Hello, World",
            media_type="application/octet-stream",
            url="http://hello.world/foo",
        )

    result = EventImporter("jid", oi, bi, vei).import_data([event1.as_dict()])
    assert result["event"]["insert"] == 1

    result = EventImporter("jid", oi, bi, vei).import_data([event2.as_dict()])
    assert result["event"]["noop"] == 1


@pytest.mark.django_db
def test_media_document():
    create_jurisdiction()
    event1 = ge()
    event2 = ge()

    for event in [event1, event2]:
        event.add_document(
            note="Presentation", url="http://example.com/presentation.pdf"
        )

    result = EventImporter("jid", oi, bi, vei).import_data([event1.as_dict()])
    assert result["event"]["insert"] == 1

    result = EventImporter("jid", oi, bi, vei).import_data([event2.as_dict()])
    assert result["event"]["noop"] == 1


@pytest.mark.django_db
def test_full_event():
    create_jurisdiction()
    george = Person.objects.create(id="gw", name="George Washington")
    o = Organization.objects.create(name="Merica", jurisdiction_id="jid")
    Membership.objects.create(person=george, organization=o)

    event = ge()

    result = EventImporter("jid", oi, bi, vei).import_data([event.as_dict()])
    assert result["event"]["insert"] == 1

    event = ge()

    result = EventImporter("jid", oi, bi, vei).import_data([event.as_dict()])
    assert result["event"]["noop"] == 1

    event = ge()
    event.location["name"] = "United States of America"
    result = EventImporter("jid", oi, bi, vei).import_data([event.as_dict()])
    assert result["event"]["update"] == 1


@pytest.mark.django_db
def test_dedupe_key_event():
    create_jurisdiction()
    create_other_jurisdiction()
    george = Person.objects.create(id="gw", name="George Washington")
    o = Organization.objects.create(name="Merica", jurisdiction_id="jid")
    Membership.objects.create(person=george, organization=o)

    event = ge()
    event.dedupe_key = "foo"

    result = EventImporter("jid", oi, bi, vei).import_data([event.as_dict()])
    assert result["event"]["insert"] == 1

    result = EventImporter("jid", oi, bi, vei).import_data([event.as_dict()])
    assert result["event"]["noop"] == 1

    event.name = ("America's Anniversary",)
    event.location["name"] = "United States of America"
    result = EventImporter("jid", oi, bi, vei).import_data([event.as_dict()])
    assert result["event"]["update"] == 1

    event.dedupe_key = "bar"
    result = EventImporter("jid", oi, bi, vei).import_data([event.as_dict()])
    assert result["event"]["insert"] == 1

    result = EventImporter("ojid", oi, bi, vei).import_data([event.as_dict()])
    assert result["event"]["insert"] == 1


# @pytest.mark.django_db
# def test_bad_event_date():
#     create_jurisdiction()
#     event = ge()
#     event.start_date = '2017'
#     pytest.raises(
#         ValueError,
#         EventImporter('jid', oi, bi, vei).import_item,
#         event.as_dict()
#     )


@pytest.mark.django_db
def test_top_level_media_event():
    create_jurisdiction()
    event1, event2 = ge(), ge()

    event1.add_media_link(
        "fireworks",
        "http://example.com/fireworks.mov",
        media_type="application/octet-stream",
    )
    event2.add_media_link(
        "fireworks",
        "http://example.com/fireworks.mov",
        media_type="application/octet-stream",
    )

    result = EventImporter("jid", oi, bi, vei).import_data([event1.as_dict()])
    assert result["event"]["insert"] == 1

    result = EventImporter("jid", oi, bi, vei).import_data([event2.as_dict()])
    assert result["event"]["noop"] == 1


@pytest.mark.django_db
def test_event_agenda_item():
    create_jurisdiction()
    event1 = ge()

    agenda = event1.add_agenda_item("first item")
    agenda["extras"] = {"one": 1, "two": [2]}

    result = EventImporter("jid", oi, bi, vei).import_data([event1.as_dict()])
    assert result["event"]["insert"] == 1

    e = Event.objects.get()
    a = e.agenda.all()[0]
    assert a.extras == {"one": 1, "two": [2]}



from .common import (
    links,
    contact_details,
    identifiers,
    other_names,
    sources,
    extras,
)
from ...data import common

schema = {
    "properties": {
        "name": {"type": "string", "minLength": 1},
        "other_names": other_names,
        "identifiers": identifiers,
        "classification": {
            "type": ["string", "null"],
            "enum": common.ORGANIZATION_CLASSIFICATIONS,
        },
        "parent_id": {"type": ["string", "null"]},
        "contact_details": contact_details,
        "links": links,
        "sources": sources,
        # added to popolo
        "jurisdiction_id": {"type": "string", "minLength": 1},
        "division_id": {"type": ["string", "null"], "minLength": 1},
        "extras": extras,
    },
    "type": "object",
}



# Generated by Django 3.0.5 on 2020-09-14 12:16

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0024_auto_20200914_1101"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="billdocumentlink",
            name="text",
        ),
        migrations.RemoveField(
            model_name="billversionlink",
            name="text",
        ),
        migrations.RemoveField(
            model_name="eventagendamedialink",
            name="text",
        ),
        migrations.RemoveField(
            model_name="eventdocumentlink",
            name="text",
        ),
        migrations.RemoveField(
            model_name="eventmedialink",
            name="text",
        ),
    ]



from .common import extras, fuzzy_date_string

schema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "minLength": 1},
        "url": {"type": "string", "minLength": 1},
        "classification": {"type": "string", "minLength": 1},  # TODO: enum
        "division_id": {"type": "string", "minLength": 1},
        "legislative_sessions": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string", "minLength": 1},
                    "type": {"type": "string", "enum": ["primary", "special"]},
                    "start_date": {
                        "type": [fuzzy_date_string, "date"],
                        "required": True,
                    },
                    "end_date": {"type": [fuzzy_date_string, "date"], "required": True},
                },
            },
        },
        "extras": extras,
    },
}



from django.db import migrations, models
import django.db.models.deletion
from ...metadata import lookup


def forwards_func(apps, schema_editor):
    # copy cities into jurisdiction ids
    Person = apps.get_model("data", "Person")

    for p in Person.objects.exclude(current_state=""):
        p.current_jurisdiction_id = lookup(abbr=p.current_state).jurisdiction_id
        p.save()


class Migration(migrations.Migration):

    dependencies = [("data", "0010_auto_20200721_1604")]

    operations = [
        migrations.AddField(
            model_name="person",
            name="current_jurisdiction",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="current_people",
                to="data.Jurisdiction",
            ),
        ),
        migrations.RunPython(forwards_func),
    ]



from ..models import State, Chamber, simple_numbered_districts

RI = State(
    name="Rhode Island",
    abbr="RI",
    capital="Providence",
    capital_tz="America/New_York",
    fips="44",
    unicameral=False,
    legislature_name="Rhode Island General Assembly",
    legislature_organization_id="ocd-organization/fce467b7-470b-41c2-be23-0ed00804b512",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/3c6b003d-47bf-44ca-a3cf-fe7a53780cc8",
    division_id="ocd-division/country:us/state:ri",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ri/government",
    url="http://www.ri.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/ba73a081-69ab-42c3-bf47-7620fec808cf",
        num_seats=75,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ri", "lower", 75
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/bea6e1af-6d6b-4a81-8040-499daf785244",
        num_seats=38,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ri", "upper", 38
        ),
    ),
)



import copy
from .base import (
    BaseModel,
    SourceMixin,
    LinkMixin,
    ContactDetailMixin,
    OtherNameMixin,
    IdentifierMixin,
)
from .schemas.post import schema as post_schema
from .schemas.person import schema as person_schema
from .schemas.membership import schema as membership_schema
from .schemas.organization import schema as org_schema
from ..utils import _make_pseudo_id
from ..exceptions import ScrapeValueError

# a copy of the org schema without sources
org_schema_no_sources = copy.deepcopy(org_schema)
org_schema_no_sources["properties"].pop("sources")


class Post(BaseModel):
    """
    A popolo-style Post
    """

    _type = "post"
    _schema = post_schema

    def __init__(
        self,
        *,
        label,
        role,
        organization_id=None,
        chamber=None,
        division_id=None,
        start_date="",
        end_date="",
        maximum_memberships=1
    ):
        super(Post, self).__init__()
        self.label = label
        self.role = role
        self.organization_id = pseudo_organization(organization_id, chamber)
        self.division_id = division_id
        self.start_date = start_date
        self.end_date = end_date
        self.maximum_memberships = maximum_memberships

    def __str__(self):
        return self.label


class Membership(BaseModel):
    """
    A popolo-style Membership.
    """

    _type = "membership"
    _schema = membership_schema

    def __init__(
        self,
        *,
        person_id,
        organization_id,
        post_id=None,
        role="",
        label="",
        start_date="",
        end_date="",
        person_name=""
    ):
        """
        Constructor for the Membership object.

        We require a person ID and organization ID, as required by the
        popolo spec. Additional arguments may be given, which match those
        defined by popolo.
        """
        super(Membership, self).__init__()
        self.person_id = person_id
        self.person_name = person_name
        self.organization_id = organization_id
        self.post_id = post_id
        self.start_date = start_date
        self.end_date = end_date
        self.role = role
        self.label = label

    def __str__(self):
        return self.person_id + " membership in " + self.organization_id


class Person(
    BaseModel,
    SourceMixin,
    ContactDetailMixin,
    LinkMixin,
    IdentifierMixin,
    OtherNameMixin,
):
    """
    Details for a Person in Popolo format.
    """

    _type = "person"
    _schema = person_schema

    def __init__(
        self,
        name,
        *,
        birth_date="",
        death_date="",
        biography="",
        image="",
        gender="",
        # specialty fields
        district=None,
        party=None,
        primary_org="",
        role="",
        start_date="",
        end_date="",
        primary_org_name=None
    ):
        super(Person, self).__init__()
        self.name = name
        self.birth_date = birth_date
        self.death_date = death_date
        self.biography = biography
        self.image = image
        self.gender = gender
        if primary_org:
            self.add_term(
                role,
                primary_org,
                district=district,
                start_date=start_date,
                end_date=end_date,
                org_name=primary_org_name,
            )
        if party:
            self.add_party(party)

    def add_membership(self, name_or_org, role="member", **kwargs):
        """
        add a membership in an organization and return the membership
        object in case there are more details to add
        """
        if isinstance(name_or_org, Organization):
            membership = Membership(
                person_id=self._id,
                person_name=self.name,
                organization_id=name_or_org._id,
                role=role,
                **kwargs
            )
        else:
            membership = Membership(
                person_id=self._id,
                person_name=self.name,
                organization_id=_make_pseudo_id(name=name_or_org),
                role=role,
                **kwargs
            )
        self._related.append(membership)
        return membership

    def add_party(self, party, **kwargs):
        membership = Membership(
            person_id=self._id,
            person_name=self.name,
            organization_id=_make_pseudo_id(classification="party", name=party),
            role="member",
            **kwargs
        )
        self._related.append(membership)

    def add_term(
        self,
        role,
        org_classification,
        *,
        district=None,
        start_date="",
        end_date="",
        label="",
        org_name=None,
        appointment=False
    ):
        if org_name:
            org_id = _make_pseudo_id(classification=org_classification, name=org_name)
        else:
            org_id = _make_pseudo_id(classification=org_classification)

        if district:
            if role:
                post_id = _make_pseudo_id(
                    label=district,
                    role=role,
                    organization__classification=org_classification,
                )
            else:
                post_id = _make_pseudo_id(
                    label=district, organization__classification=org_classification
                )
        elif appointment:
            post_id = _make_pseudo_id(
                role=role, organization__classification=org_classification
            )
        else:
            post_id = None
        membership = Membership(
            person_id=self._id,
            person_name=self.name,
            organization_id=org_id,
            post_id=post_id,
            role=role,
            start_date=start_date,
            end_date=end_date,
            label=label,
        )
        self._related.append(membership)
        return membership

    def __str__(self):
        return self.name


class Organization(BaseModel, SourceMixin, LinkMixin):
    """
    A single popolo-style Organization
    """

    _type = "organization"
    _schema = org_schema

    def __init__(self, name, *, classification="", parent_id=None, chamber=None):
        """
        Constructor for the Organization object.
        """
        super(Organization, self).__init__()
        self.name = name
        self.classification = classification
        self.parent_id = pseudo_organization(parent_id, chamber)

    def __str__(self):
        return self.name

    def validate(self):
        schema = None
        # these are implicitly declared & do not require sources
        if self.classification in (
            "party",
            "legislature",
            "upper",
            "lower",
            "executive",
        ):
            schema = org_schema_no_sources
        return super(Organization, self).validate(schema=schema)

    def add_post(self, label, role, **kwargs):
        post = Post(label=label, role=role, organization_id=self._id, **kwargs)
        self._related.append(post)
        return post

    def add_member(self, name_or_person, role="member", **kwargs):
        if isinstance(name_or_person, Person):
            membership = Membership(
                person_id=name_or_person._id,
                person_name=name_or_person.name,
                organization_id=self._id,
                role=role,
                **kwargs
            )
        else:
            membership = Membership(
                person_id=_make_pseudo_id(name=name_or_person),
                person_name=name_or_person,
                organization_id=self._id,
                role=role,
                **kwargs
            )
        self._related.append(membership)
        return membership


def pseudo_organization(organization, classification, default=None):
    """ helper for setting an appropriate ID for organizations """
    if organization and classification:
        raise ScrapeValueError("cannot specify both classification and organization")
    elif classification:
        return _make_pseudo_id(classification=classification)
    elif organization:
        if isinstance(organization, Organization):
            return organization._id
        elif isinstance(organization, str):
            return organization
        else:
            return _make_pseudo_id(**organization)
    elif default is not None:
        return _make_pseudo_id(classification=default)
    else:
        return None



# flake8: noqa
from .jurisdiction import JurisdictionImporter
from .organizations import OrganizationImporter
from .bills import BillImporter
from .vote_events import VoteEventImporter
from .events import EventImporter



import pytest
from openstates.data.models import Jurisdiction, Division, Organization, Person, Post
from django.core.exceptions import ValidationError


def test_division_subtypes_from_id():

    # simplest case
    assert Division.subtypes_from_id("ocd-division/country:us") == (
        {"country": "us"},
        1,
    )

    # ocd-divison part is optional
    assert Division.subtypes_from_id("country:us/state:ak/county:wild") == (
        {
            "country": "us",
            "subtype1": "state",
            "subid1": "ak",
            "subtype2": "county",
            "subid2": "wild",
        },
        3,
    )

    # country is not optional
    with pytest.raises(ValueError):
        Division.subtypes_from_id("state:nc/city:raleigh")


@pytest.mark.django_db
def test_division_create():
    division_id = "ocd-division/country:us/state:ak/county:wild"
    d = Division.objects.create(id=division_id, name="Wild County")
    assert d.country == "us"
    assert d.subtype1 == "state"
    assert d.subid1 == "ak"
    assert d.subtype2 == "county"
    assert d.subid2 == "wild"
    assert division_id in str(d)


@pytest.mark.django_db
def test_division_children_of():
    Division.objects.create("ocd-division/country:us", name="US")
    ak = Division.objects.create("ocd-division/country:us/state:ak", name="Alaska")
    Division.objects.create("ocd-division/country:us/state:ak/county:wild", name="Wild")
    Division.objects.create("ocd-division/country:us/state:ak/county:mild", name="Mild")
    Division.objects.create(
        "ocd-division/country:us/state:ak/county:wild/place:a", name="A"
    )
    Division.objects.create(
        "ocd-division/country:us/state:ak/county:wild/place:b", name="B"
    )
    Division.objects.create(
        "ocd-division/country:us/state:ak/county:wild/school:a", name="A"
    )
    Division.objects.create(
        "ocd-division/country:us/state:ak/county:mild/place:a", name="A"
    )
    Division.objects.create(
        "ocd-division/country:us/state:ak/county:mild/place:a/x:y", name="A"
    )

    # simplest ==
    assert Division.objects.children_of("ocd-division/country:us")[0].id == ak.id

    # 3 divisions within wild county
    assert (
        Division.objects.children_of(
            "ocd-division/country:us/state:ak/county:wild"
        ).count()
        == 3
    )

    # only one school in wild county
    assert (
        Division.objects.children_of(
            "ocd-division/country:us/state:ak/county:wild", subtype="school"
        ).count()
        == 1
    )

    # 6 divisions beneath alaska up to 2 levels
    assert (
        Division.objects.children_of(
            "ocd-division/country:us/state:ak", depth=2
        ).count()
        == 6
    )

    # 7 divisions beneath alaska up to 3 levels
    assert (
        Division.objects.children_of(
            "ocd-division/country:us/state:ak", depth=3
        ).count()
        == 7
    )


@pytest.mark.django_db
def test_ocdid_default():
    o = Organization.objects.create(name="test org")
    assert o.id.startswith("ocd-organization/")
    assert o.pk == o.id
    p = Person.objects.create(name="test person")
    assert p.id.startswith("ocd-person/")


@pytest.mark.django_db
def test_ocdid_default_nondup():
    """ ensure that defaults actually vary """
    p1 = Person(name="test person 1")
    p2 = Person(name="test person 2")
    assert p1.id != p2.id


@pytest.mark.django_db
def test_ocdid_validation_jurisdiction():
    # this fails
    with pytest.raises(ValidationError):
        j = Jurisdiction(
            name="test juris",
            id="ocd-division/country:us/test:something/else",
            url="http://example.com",
        )
        j.full_clean(exclude=["division"])

    # this succeeds
    j = Jurisdiction(
        name="test juris",
        id="ocd-jurisdiction/country:us/test:something/else",
        url="http://example.com",
    )
    j.full_clean(exclude=["division"])

    assert str(j) == j.name


@pytest.mark.django_db
def test_ocdid_validation_other():
    # this test should handle everything that isn't a jurisdiction

    # this succeeds
    o = Organization(name="test org")
    o.full_clean(exclude=["parent", "jurisdiction"])

    # this raises
    with pytest.raises(ValidationError):
        o = Organization(name="this is a test", id="ocd-organization/3")
        o.full_clean(exclude=["parent", "jurisdiction"])


@pytest.mark.django_db
def test_organization_get_parents():
    o1 = Organization.objects.create(
        name="National Organization on Bread-and-Cheese Products"
    )
    o2 = Organization.objects.create(name="Committee on Pizza", parent=o1)
    o3 = Organization.objects.create(name="Subcommittee on Sicilian Pizza", parent=o2)

    assert list(o3.get_parents()) == [o2, o1]


@pytest.mark.django_db
def test_organization_str():
    o = Organization.objects.create(name="test org")
    assert "test org" in str(o)


@pytest.mark.django_db
def test_organization_post():
    o = Organization.objects.create(name="test org")
    o.posts.create(label="CEO")

    assert "CEO" in str(o.posts.all()[0])
    assert "test org" in str(o.posts.all()[0])


@pytest.mark.django_db
def test_organization_membership():
    o = Organization.objects.create(name="test org")
    p = Person.objects.create(name="test person")

    o.memberships.create(person=p, role="CEO")

    assert "CEO" in str(o.memberships.all()[0])
    assert "test person" in str(o.memberships.all()[0])
    assert "test org" in str(o.memberships.all()[0])
    assert len(o.get_current_members()) == 1
    assert len(Person.objects.member_of("test org")) == 1
    assert len(Person.objects.member_of(o.id)) == 1


@pytest.mark.django_db
def test_member_of_with_post():
    o = Organization.objects.create(name="The Org")
    p = Post.objects.create(organization=o, label="1")
    cur = Person.objects.create(name="current")
    prev = Person.objects.create(name="previous")
    o.memberships.create(person=cur, post=p)
    o.memberships.create(person=prev, post=p, end_date="2019-01-01")

    assert len(o.get_current_members()) == 1
    assert len(Person.objects.member_of("The Org")) == 1
    assert len(Person.objects.member_of("The Org", post="1")) == 1
    assert len(Person.objects.member_of("The Org", current_only=False)) == 2


@pytest.mark.django_db
def test_member_of_prior_role_conflict():
    # test for https://github.com/openstates/openstates.org/issues/304
    o = Organization.objects.create(name="The Org")
    p = Post.objects.create(organization=o, label="1")
    p2 = Post.objects.create(organization=o, label="2")
    cur = Person.objects.create(name="current")
    prev = Person.objects.create(name="previous")
    o.memberships.create(person=cur, post=p)
    o.memberships.create(person=prev, post=p, end_date="2019-01-01")
    o.memberships.create(person=prev, post=p2)

    assert len(o.get_current_members()) == 2
    assert len(Person.objects.member_of("The Org")) == 2
    assert len(Person.objects.member_of("The Org", post="1")) == 1
    assert len(Person.objects.member_of("The Org", post="1", current_only=False)) == 2


@pytest.mark.django_db
def test_person_query_active_members(senator):
    assert Person.objects.count() == 1
    people = Person.objects.active()
    assert len(people) == 1
    assert people[0].name == "Willy Worm"


@pytest.mark.django_db
def test_person_query_current_with_roles(senator, django_assert_num_queries):
    senate = senator.memberships.filter(organization__classification="upper")[
        0
    ].organization

    # prefetch grabs membership, org, post too
    with django_assert_num_queries(4):
        willy = Person.objects.current_legislators_with_roles([senate])[0]
    assert willy.name == "Willy Worm"


@pytest.mark.django_db
def test_person_search(senator, person):
    assert Person.objects.search("Worm").count() == 1
    assert Person.objects.search("Worm", state="mo").count() == 1
    assert Person.objects.search("Worm", state="wy").count() == 0
    assert Person.objects.search("Pig").count() == 0


@pytest.mark.django_db
def test_person_str(person):
    assert person.name in str(person)


@pytest.mark.django_db
def test_legislative_session_str(legislative_session):
    assert legislative_session.name in str(legislative_session)


@pytest.mark.django_db
def test_vote_event_str(vote_event):
    assert vote_event.motion_text in str(vote_event)
    # test adding identifier and alternate string repr
    vote_event.identifier = "Roll Call #2372"
    vote_event.save()
    assert "Roll Call #2372" in str(vote_event)


@pytest.mark.django_db
def test_vote_event_count(vote_event):
    vote_event.counts.create(option="yes", value=36)
    assert "yes" in str(vote_event.counts.all()[0])


@pytest.mark.django_db
def test_vote_event_vote(vote_event):
    p = Person.objects.create(name="Maria Chappelle-Nadal")
    vote_event.votes.create(option="yes", voter_name="Maria Chappelle-Nadal", voter=p)
    assert "Maria Chappelle-Nadal" in str(vote_event.votes.all()[0])


@pytest.mark.django_db
def test_bill_str(bill):
    assert bill.identifier in str(bill)


@pytest.mark.django_db
def test_bill_abstract(bill):
    bill.abstracts.create(
        abstract="This is the Senate's health care bill. The bill started off "
        "with text regarding an unrelated matter but the Senate is "
        "co-opted this bill as a vehicle for passage of their reform and "
        "changed the text in whole to the health care bill. They do this "
        "because the Constitution requires all revenue bills to start in "
        "the House, and their health reform plan involves revenue. So "
        "they have chosen to work off of a bill that started in the "
        "House, even if that bill is unrelated.",
        note="Library of Congress Summary",
    )
    assert bill.identifier in str(bill.abstracts.all()[0])


@pytest.mark.django_db
def test_bill_title(bill):
    o_t = bill.other_titles.create(title="Affordable Care Act")
    assert o_t.title in str(o_t)
    assert bill.identifier in str(o_t)


@pytest.mark.django_db
def test_bill_sponsorship(bill):
    spon = bill.sponsorships.create(classification="sponsor", name="Nancy Pelosi")
    assert spon.name in str(spon)
    assert bill.identifier in str(spon)


@pytest.mark.django_db
def test_bill_identifier(bill):
    bill.other_identifiers.create(identifier="1001")
    assert "1001" == bill.other_identifiers.all()[0].identifier


@pytest.mark.django_db
def test_related_bill(bill):
    r_b = bill.related_bills.create(
        legislative_session=bill.legislative_session,
        identifier="SB 22",
        relation_type="companion",
    )
    assert r_b.relation_type in str(r_b)
    assert bill.identifier in str(r_b)


@pytest.mark.django_db
def test_bill_action(bill):
    o = Organization.objects.create(name="Missouri State Senate")
    a = bill.actions.create(
        organization=o,
        description="Third Reading and Final Passage",
        date="2016-02-16",
        order=1,
    )
    assert a.date in str(a)
    assert bill.identifier in str(a)

    # test adding related entity to bill action
    p = Person.objects.create(name="Maria Chappelle-Nadal")
    a.related_entities.create(person=p, name=p.name, entity_type="person")
    assert p.name in str(a.related_entities.all()[0])


@pytest.mark.django_db
def test_bill_version_with_links(bill):
    v = bill.versions.create(note="Engrossed", date="2017-03-15")
    assert v.date in str(v)
    assert bill.identifier in str(v)

    # test adding link bill version
    v.links.create(
        url="http://committee.example.com/billversion1.pdf",
        media_type="application/pdf",
    )
    assert "http://committee.example.com/billversion1.pdf" in str(v.links.all()[0])
    assert bill.identifier in str(v.links.all()[0])


@pytest.mark.django_db
def test_bill_document_with_links(bill):
    doc = bill.documents.create(note="Fiscal Note", date="2017-03-01")
    assert bill.identifier in str(doc)

    # test adding link bill version
    doc.links.create(
        url="http://committee.example.com/bill_document.pdf",
        media_type="application/pdf",
    )
    assert "http://committee.example.com/bill_document.pdf" in str(doc.links.all()[0])
    assert bill.identifier in str(doc.links.all()[0])


@pytest.mark.django_db
def test_event_str(event):
    assert event.name in str(event)
    assert event.start_date in str(event)


@pytest.mark.django_db
def test_event_location_str(event_location):
    assert event_location.name in str(event_location)


@pytest.mark.django_db
def test_event_participant_person(event):
    p = Person.objects.create(name="Andrew Tobin")
    e_p = event.participants.create(
        name=p.name, person=p, entity_type="person", note="Speaker"
    )
    assert e_p.name in str(e_p)
    assert e_p.name in e_p.entity_name
    assert e_p.entity_id
    # remove entity_type value and re-try
    e_p.entity_type = ""
    assert e_p.name in e_p.entity_name
    assert e_p.entity_id is None


@pytest.mark.django_db
def test_event_participant_organization(event):
    o = Organization.objects.create(name="Committee on Energy")
    e_p = event.participants.create(
        name=o.name, organization=o, entity_type="organization", note="Host Committee"
    )
    assert e_p.name in str(e_p)
    assert e_p.name in e_p.entity_name
    assert e_p.entity_id
    # remove entity_type value and re-try
    e_p.entity_type = ""
    assert e_p.name in e_p.entity_name
    assert e_p.entity_id is None


@pytest.mark.django_db
def test_event_link(event):
    event.links.create(note="EPA Website", url="http://www.epa.gov/")
    assert "http://www.epa.gov/" in str(event.links.all()[0])


@pytest.mark.django_db
def test_event_media_w_links(event):
    # test adding media to event
    e_m = event.media.create(
        note="Recording of the meeting", date="2014-04-12", offset="19"
    )
    assert "Recording of the meeting" in str(e_m)

    # test adding link event media
    e_m.links.create(media_type="video/webm", url="http://example.com/video.webm")
    assert "http://example.com/video.webm" in str(e_m.links.all()[0])


@pytest.mark.django_db
def test_event_document_w_links(event):
    # test adding document to event
    e_d = event.documents.create(date="2014-04-12", note="Agenda")
    assert "Agenda" in str(e_d)
    assert event.name in str(e_d)

    # test adding link to event document
    e_d.links.create(
        url="http://committee.example.com/agenda.pdf", media_type="application/pdf"
    )
    assert "http://committee.example.com/agenda.pdf" in str(e_d.links.all()[0])
    assert e_d.note in str(e_d.links.all()[0])


@pytest.mark.django_db
def test_event_agenda(event, vote_event, bill):
    # test adding agenda item to event
    e_a = event.agenda.create(
        description="Presentation by Director Henry Darwin, Arizona Department "
        "of Environmental Quality, regarding the Environmental "
        "Protection Agency (EPA) Clean Power Plan proposed rule",
        order=2,
        subjects=["epa", "green energy", "environmental issues"],
    )
    assert event.name in str(e_a)

    # test adding media to event agenda item
    e_a_med = e_a.media.create(note="Recording Darwin presentation", date="2014-04-12")
    assert "Recording Darwin presentation" in str(e_a_med)

    # test adding link to event agenda item media
    e_a_med.links.create(media_type="video/mp4", url="http://example.com/video.mp4")
    assert "http://example.com/video.mp4" in str(e_a_med.links.all()[0])

    # test adding related entities to event agenda
    e_a.related_entities.create(bill=bill, entity_type="bill", name=bill.identifier)
    e_a.related_entities.create(
        vote_event=vote_event, entity_type="vote", name=vote_event.identifier
    )
    for r_e in e_a.related_entities.all():
        assert r_e.name in str(r_e)
        assert r_e.name in r_e.entity_name
        assert r_e.entity_id
        # remove entity_type value and re-try
        r_e.entity_type = ""
        assert r_e.name in r_e.entity_name
        assert r_e.entity_id is None



import typing
from ..models import State, Chamber, District
from us import STATES


seats_by_state = {
    "AL": 7,
    "AK": 1,
    "AZ": 9,
    "AR": 4,
    "CA": 53,
    "CO": 7,
    "CT": 5,
    "DE": 1,
    "FL": 27,
    "GA": 14,
    "HI": 2,
    "ID": 2,
    "IL": 18,
    "IN": 9,
    "IA": 4,
    "KS": 4,
    "KY": 6,
    "LA": 6,
    "ME": 2,
    "MD": 8,
    "MA": 9,
    "MI": 14,
    "MN": 8,
    "MS": 4,
    "MO": 8,
    "MT": 1,
    "NE": 3,
    "NV": 4,
    "NH": 2,
    "NJ": 12,
    "NM": 3,
    "NY": 27,
    "NC": 13,
    "ND": 1,
    "OH": 16,
    "OK": 5,
    "OR": 5,
    "PA": 18,
    "RI": 2,
    "SC": 7,
    "SD": 1,
    "TN": 9,
    "TX": 36,
    "UT": 4,
    "VT": 1,
    "VA": 11,
    "WA": 10,
    "WV": 3,
    "WI": 8,
    "WY": 1,
}


def us_house_districts() -> typing.List[District]:
    districts = []

    # TODO: add other territories
    territories = [
        District(
            "DC-AL",
            "lower",
            "ocd-division/country:us/district:dc",
            title_override="Delegate",
        ),
        District(
            "PR-AL",
            "lower",
            "ocd-division/country:us/territory:pr",
            title_override="Resident Commissioner",
        ),
    ]
    districts += territories

    for abbr, num_districts in seats_by_state.items():
        if num_districts == 1:
            districts.append(
                District(
                    abbr + "-AL",
                    "lower",
                    f"ocd-division/country:us/state:{abbr.lower()}/cd:at-large",
                )
            )
        else:
            districts.extend(
                [
                    District(
                        f"{abbr}-{n}",
                        "lower",
                        f"ocd-division/country:us/state:{abbr.lower()}/cd:{n}",
                    )
                    for n in range(1, num_districts + 1)
                ]
            )
    return districts


def us_senate_districts() -> typing.List[District]:
    return [
        District(s.name, "upper", f"ocd-division/country:us/state:{s.abbr.lower()}", 2)
        for s in STATES
    ]


US = State(
    name="United States",
    abbr="US",
    capital="Washington DC",
    capital_tz="America/New_York",
    fips="00",
    unicameral=False,
    legislature_name="US Congress",
    legislature_organization_id="ocd-organization/98590d80-8f8a-57fa-bdfc-55d778fbd88a",
    executive_name="President of the United States",
    executive_organization_id="ocd-organization/16b52117-31a3-5f05-96a6-fd580a4f9f4d",
    division_id="ocd-division/country:us",
    jurisdiction_id="ocd-jurisdiction/country:us/government",
    url="https://usa.gov",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/24af4233-d9b5-5933-91b2-51d29f721037",
        num_seats=437,
        title="Representative",
        districts=us_house_districts(),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/072da2ce-df81-52c3-9cc8-323e208cdf10",
        num_seats=100,
        title="Senator",
        districts=us_senate_districts(),
    ),
)



from django.db import models


class DivisionManager(models.Manager):
    def children_of(self, division_id, subtype=None, depth=1):
        query, n = Division.subtypes_from_id(division_id)
        q_objects = []

        # only get children
        if subtype:
            query["subtype{0}".format(n)] = subtype
        else:
            q_objects.append(~models.Q(**{"subtype{0}".format(n): ""}))
        q_objects.append(~models.Q(**{"subid{0}".format(n): ""}))

        # allow for depth wildcards
        n += depth

        # ensure final field is null
        q_objects.append(models.Q(**{"subtype{0}".format(n): ""}))
        q_objects.append(models.Q(**{"subid{0}".format(n): ""}))

        return self.filter(*q_objects, **query)

    def create(self, id, name, redirect=None):
        return super(DivisionManager, self).create(
            id=id, name=name, redirect=redirect, **Division.subtypes_from_id(id)[0]
        )


class Division(models.Model):
    """
    A political geography, which may have multiple boundaries over its lifetime.

    Types of divisions include, among others:
    * Governmental jurisdiction - A division that a government has jurisdiction over.
    (e.g. North Carolina)
    * Political district - A division that elects a representative to a legislature.
    (e.g. North Carolina Congressional District 4)
    * Service zone - An area to which a government provides a service.
    (e.g. Washington DC Police District 105)
    """

    objects = DivisionManager()

    id = models.CharField(max_length=300, primary_key=True)
    name = models.CharField(max_length=300, help_text="The name of the division.")
    # cascade is SET_NULL, will un-redirect if deletion happens
    redirect = models.ForeignKey("self", null=True, on_delete=models.SET_NULL)
    country = models.CharField(
        max_length=2,
        help_text="An ISO-3166-1 alpha-2 code identifying the county where this division is found.",
    )

    # up to 7 pieces of the id that are searchable
    subtype1 = models.CharField(
        max_length=50,
        blank=True,
        help_text="The first subtype in the unique identifier.",
    )
    subid1 = models.CharField(
        max_length=100,
        blank=True,
        help_text="The first subidentifier in the unique identifer.",
    )
    subtype2 = models.CharField(
        max_length=50,
        blank=True,
        help_text="The second subtype in the unique identifier.",
    )
    subid2 = models.CharField(
        max_length=100,
        blank=True,
        help_text="The second subidentifier in the unique identifer.",
    )
    subtype3 = models.CharField(
        max_length=50,
        blank=True,
        help_text="The third subtype in the unique identifier.",
    )
    subid3 = models.CharField(
        max_length=100,
        blank=True,
        help_text="The third subidentifier in the unique identifer.",
    )
    subtype4 = models.CharField(
        max_length=50,
        blank=True,
        help_text="The fourth subtype in the unique identifier.",
    )
    subid4 = models.CharField(
        max_length=100,
        blank=True,
        help_text="The fourth subidentifier in the unique identifer.",
    )
    subtype5 = models.CharField(
        max_length=50,
        blank=True,
        help_text="The fifth subtype in the unique identifier.",
    )
    subid5 = models.CharField(
        max_length=100,
        blank=True,
        help_text="The fifth subidentifier in the unique identifer.",
    )
    subtype6 = models.CharField(
        max_length=50,
        blank=True,
        help_text="The sixth subtype in the unique identifier.",
    )
    subid6 = models.CharField(
        max_length=100,
        blank=True,
        help_text="The sixth subidentifier in the unique identifer.",
    )
    subtype7 = models.CharField(
        max_length=50,
        blank=True,
        help_text="The seventh subtype in the unique identifier.",
    )
    subid7 = models.CharField(
        max_length=100,
        blank=True,
        help_text="The seventh subidentifier in the unique identifer.",
    )

    class Meta:
        db_table = "opencivicdata_division"

    def __str__(self):
        return "{0} ({1})".format(self.name, self.id)

    @staticmethod
    def subtypes_from_id(division_id):
        pieces = [piece.split(":", 1) for piece in division_id.split("/")]
        fields = {}

        # if it included the ocd-division bit, pop it off
        if pieces[0] == ["ocd-division"]:
            pieces.pop(0)

        if pieces[0][0] != "country":
            raise ValueError("OCD id must start with country")

        fields["country"] = pieces[0][1]

        # add the remaining pieces
        n = 1
        for stype, subid in pieces[1:]:
            fields["subtype{0}".format(n)] = stype
            fields["subid{0}".format(n)] = subid
            n += 1

        return fields, n



import click
from django.db import transaction
from .. import metadata
from ..utils.django import init_django


def update_bill_fields_for_state(abbr):
    from ..data.models import Bill
    from ..importers.computed_fields import update_bill_fields

    state = metadata.lookup(abbr=abbr)

    with transaction.atomic():
        bills = Bill.objects.filter(
            legislative_session__jurisdiction=state.jurisdiction_id
        )

        with click.progressbar(bills, label=f"updating {abbr} bills") as bills_p:
            for bill in bills_p:
                update_bill_fields(bill, save=True)


@click.command()
@click.argument("abbrs", nargs=-1)
def main(abbrs):
    """ updates computed fields """
    init_django()
    if not abbrs:
        abbrs = metadata.STATES_BY_ABBR.keys()
    for abbr in abbrs:
        update_bill_fields_for_state(abbr)



from ..models import State, Chamber, simple_numbered_districts

GA = State(
    name="Georgia",
    abbr="GA",
    capital="Atlanta",
    capital_tz="America/New_York",
    fips="13",
    unicameral=False,
    legislature_name="Georgia General Assembly",
    legislature_organization_id="ocd-organization/1186f244-c9f8-46de-a453-c58d6de050d7",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/f35a01bf-5beb-5437-9842-e5ba89734908",
    division_id="ocd-division/country:us/state:ga",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ga/government",
    url="http://www.legis.ga.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/f94ae35f-f161-4599-b603-e7737dc8c93f",
        num_seats=180,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ga", "lower", 180
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/3b9777f2-c2fb-48e1-83bf-66fc7a17492e",
        num_seats=56,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ga", "upper", 56
        ),
    ),
)



from django.contrib import admin
from .base import ModelAdmin, ReadOnlyTabularInline
from .. import models


class VoteCountInline(ReadOnlyTabularInline):
    model = models.VoteCount
    fields = readonly_fields = ("option", "value")


class PersonVoteInline(ReadOnlyTabularInline):
    model = models.PersonVote
    fields = readonly_fields = ("voter", "voter_name", "option")


class VoteSourceInline(ReadOnlyTabularInline):
    model = models.VoteSource
    fields = readonly_fields = ("url", "note")


@admin.register(models.VoteEvent)
class VoteEventAdmin(ModelAdmin):
    readonly_fields = (
        "bill",
        "organization",
        "legislative_session",
        "id",
        "identifier",
        "motion_text",
        "extras",
    )
    fields = readonly_fields + (
        "result",
        "motion_classification",
        "start_date",
        "end_date",
    )

    list_selected_related = (
        "sources",
        "legislative_session",
        "legislative_session__jurisdiction",
        "counts",
    )

    def get_jurisdiction_name(self, obj):
        return obj.legislative_session.jurisdiction.name

    get_jurisdiction_name.short_description = "Jurisdiction"

    def get_vote_tally(self, obj):
        yes = no = other = 0
        for vc in obj.counts.all():
            if vc.option == "yes":
                yes = vc.value
            elif vc.option == "no":
                no = vc.value
            else:
                other += vc.value
        return "{}-{}-{}".format(yes, no, other)

    get_vote_tally.short_description = "Vote Tally"

    list_display = ("get_jurisdiction_name", "identifier", "bill", "get_vote_tally")

    list_filter = ("legislative_session__jurisdiction__name",)

    inlines = [VoteCountInline, PersonVoteInline, VoteSourceInline]



from collections import defaultdict
from openstates.scrape import Jurisdiction, Organization, JurisdictionScraper
from openstates.exceptions import ScrapeValueError
import pytest


class FakeJurisdiction(Jurisdiction):
    division_id = "ocd-division/test"
    classification = "government"
    name = "Test"
    url = "http://example.com"

    def get_organizations(self):
        parent = Organization("Congress", classification="legislature")
        yield parent
        yield Organization("House", classification="lower", parent_id=parent)
        yield Organization("Senate", classification="upper", parent_id=parent)


def test_basics():
    # id property and string
    j = FakeJurisdiction()
    assert j.jurisdiction_id == "ocd-jurisdiction/test/government"
    assert j.name in str(j)


def test_as_dict():
    j = FakeJurisdiction()
    d = j.as_dict()

    assert d["_id"] == j.jurisdiction_id
    assert d["name"] == j.name
    assert d["url"] == j.url
    assert d["legislative_sessions"] == []


def test_jurisdiction_unicam_scrape():
    class UnicameralJurisdiction(Jurisdiction):
        jurisdiction_id = "unicam"
        name = "Unicameral"
        url = "http://example.com"

        def get_organizations(self):
            yield Organization("Unicameral Legislature", classification="legislature")

    j = UnicameralJurisdiction()
    js = JurisdictionScraper(j, "/tmp/")
    objects = list(js.scrape())

    # two objects, first is the Jurisdiction
    assert len(objects) == 2
    assert objects[0] == j

    # ensure we made a single legislature org
    assert isinstance(objects[1], Organization)
    assert objects[1].classification == "legislature"


def test_jurisdiction_bicameral_scrape():
    j = FakeJurisdiction()
    js = JurisdictionScraper(j, "/tmp/")
    objects = list(js.scrape())
    obj_names = set()
    obj_types = defaultdict(int)

    for o in objects:
        obj_names.add(o.name)
        obj_types[type(o)] += 1

    # ensure Jurisdiction and 5 organizations were found
    assert obj_names == {"Test", "Congress", "House", "Senate"}
    assert obj_types[FakeJurisdiction] == 1
    assert obj_types[Organization] == 3


def test_jurisdiction_validate_sessions():
    j = FakeJurisdiction()
    j.legislative_sessions = [
        {
            "_scraped_name": "s1",
            "classification": "primary",
            "name": "First Session",
            "start_date": "2020-01-01",
        }
    ]
    with pytest.raises(ScrapeValueError):
        j.validate()
    j.legislative_sessions[0]["end_date"] = "2020-02-01"
    j.validate()



# Generated by Django 3.1.7 on 2021-04-05 17:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("reports", "0001_initial"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Identifier",
        ),
    ]



from ..models import State, Chamber, simple_numbered_districts

MI = State(
    name="Michigan",
    abbr="MI",
    capital="Lansing",
    capital_tz="America/New_York",
    fips="26",
    unicameral=False,
    legislature_name="Michigan Legislature",
    legislature_organization_id="ocd-organization/c70eb47a-43c1-4aa1-95a6-2aea30da8e82",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/1435164b-be3c-56fd-b426-678565c853cd",
    division_id="ocd-division/country:us/state:mi",
    jurisdiction_id="ocd-jurisdiction/country:us/state:mi/government",
    url="http://www.legislature.mi.gov",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/d387cf69-631c-4013-bb32-4a00c848fcc4",
        num_seats=110,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:mi", "lower", 110
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/7cf9891c-6c1f-45ec-b24c-7aa9be45e9cc",
        num_seats=38,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:mi", "upper", 38
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

OR = State(
    name="Oregon",
    abbr="OR",
    capital="Salem",
    capital_tz="America/Los_Angeles",
    fips="41",
    unicameral=False,
    legislature_name="Oregon Legislative Assembly",
    legislature_organization_id="ocd-organization/5baae940-2a55-49f9-8e8b-7d321ac539b4",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/6490d24e-7708-5ece-be11-9c7b82fc8920",
    division_id="ocd-division/country:us/state:or",
    jurisdiction_id="ocd-jurisdiction/country:us/state:or/government",
    url="https://olis.leg.state.or.us",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/9b08e1e9-5a70-48c1-983d-508a3281c378",
        num_seats=60,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:or", "lower", 60
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/bfb1aaad-4896-4399-b12f-64b8b4ab1b30",
        num_seats=30,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:or", "upper", 30
        ),
    ),
)



import pytest
from unittest import mock
from openstates.scrape import Person, Organization, Bill, Jurisdiction
from openstates.scrape.base import Scraper, ScrapeError, BaseBillScraper


class FakeJurisdiction(Jurisdiction):
    jurisdiction_id = "jurisdiction"


juris = FakeJurisdiction()


def test_save_object_basics():
    # ensure that save object dumps a file
    s = Scraper(juris, "/tmp/")
    p = Person("Michael Jordan")
    p.add_source("http://example.com")

    with mock.patch("json.dump") as json_dump:
        s.save_object(p)

    # ensure object is saved in right place
    filename = "person_" + p._id + ".json"
    assert filename in s.output_names["person"]
    json_dump.assert_called_once_with(p.as_dict(), mock.ANY, cls=mock.ANY)


def test_save_object_invalid():
    s = Scraper(juris, "/tmp/")
    p = Person("Michael Jordan")
    # no source, won't validate

    with pytest.raises(ValueError):
        s.save_object(p)


def test_save_related():
    s = Scraper(juris, "/tmp/")
    p = Person("Michael Jordan")
    p.add_source("http://example.com")
    o = Organization("Chicago Bulls", classification="committee")
    o.add_source("http://example.com")
    p._related.append(o)

    with mock.patch("json.dump") as json_dump:
        s.save_object(p)

    assert json_dump.mock_calls == [
        mock.call(p.as_dict(), mock.ANY, cls=mock.ANY),
        mock.call(o.as_dict(), mock.ANY, cls=mock.ANY),
    ]


def test_simple_scrape():
    class FakeScraper(Scraper):
        def scrape(self):
            p = Person("Michael Jordan")
            p.add_source("http://example.com")
            yield p

    with mock.patch("json.dump") as json_dump:
        record = FakeScraper(juris, "/tmp/").do_scrape()

    assert len(json_dump.mock_calls) == 1
    assert record["objects"]["person"] == 1
    assert record["end"] > record["start"]
    assert record["skipped"] == 0


def test_double_iter():
    """ tests that scrapers that yield iterables work OK """

    class IterScraper(Scraper):
        def scrape(self):
            yield self.scrape_people()

        def scrape_people(self):
            p = Person("Michael Jordan")
            p.add_source("http://example.com")
            yield p

    with mock.patch("json.dump") as json_dump:
        record = IterScraper(juris, "/tmp/").do_scrape()

    assert len(json_dump.mock_calls) == 1
    assert record["objects"]["person"] == 1


def test_no_objects():
    class NullScraper(Scraper):
        def scrape(self):
            pass

    with pytest.raises(ScrapeError):
        NullScraper(juris, "/tmp/", fastmode=True).do_scrape()


def test_no_scrape():
    class NonScraper(Scraper):
        pass

    with pytest.raises(NotImplementedError):
        NonScraper(juris, "/tmp/").do_scrape()


def test_bill_scraper():
    class BillScraper(BaseBillScraper):
        def get_bill_ids(self):
            yield "1", {"extra": "param"}
            yield "2", {}

        def get_bill(self, bill_id, **kwargs):
            if bill_id == "1":
                assert kwargs == {"extra": "param"}
                raise self.ContinueScraping
            else:
                assert bill_id == "2"
                assert kwargs == {}
                b = Bill("1", self.legislative_session, "title")
                b.add_source("http://example.com")
                return b

    bs = BillScraper(juris, "/tmp/")
    with mock.patch("json.dump") as json_dump:
        record = bs.do_scrape(legislative_session="2020")

    assert len(json_dump.mock_calls) == 1
    assert record["objects"]["bill"] == 1
    assert record["skipped"] == 1


def test_whitespace_is_stripped():
    s = Scraper(juris, "/tmp/")
    b = Bill(" HB 11", "2020", " a short title     ")
    b.subject = [" one", "two ", "   three "]
    b.add_source("https://example.com/     ")

    s.save_object(b)

    # the simple cases, and nested lists / objects
    assert b.identifier == "HB 11"
    assert b.title == "a short title"
    assert b.sources[0]["url"] == "https://example.com/"
    # subject got sorted by pre_save
    assert b.subject == ["one", "three", "two"]



import pytest
from openstates.scrape.schemas.person import schema
from openstates.scrape.base import (
    BaseModel,
    SourceMixin,
    ContactDetailMixin,
    LinkMixin,
    AssociatedLinkMixin,
    OtherNameMixin,
    IdentifierMixin,
)


class GenericModel(
    BaseModel,
    SourceMixin,
    ContactDetailMixin,
    LinkMixin,
    AssociatedLinkMixin,
    OtherNameMixin,
    IdentifierMixin,
):
    """ a generic model used for testing the base and mixins """

    _type = "generic"
    _schema = schema

    def __init__(self):
        super(GenericModel, self).__init__()
        self._associated = []


def test_init_id():
    m = GenericModel()
    assert len(m._id) == 36


def test_as_dict():
    m = GenericModel()
    assert m.as_dict()["_id"] == m._id


def test_setattr():
    m = GenericModel()

    with pytest.raises(ValueError):
        m.some_random_key = 3

    # and no error raised since this is a valid key
    m._id = "new id"


def test_add_source():
    m = GenericModel()
    m.add_source("http://example.com/1")
    m.add_source("http://example.com/2", note="xyz")
    assert m.sources == [
        {"url": "http://example.com/1", "note": ""},
        {"url": "http://example.com/2", "note": "xyz"},
    ]


def test_add_contact_detail():
    m = GenericModel()
    m.add_contact_detail(type="fax", value="111-222-3333", note="office")
    assert m.contact_details == [
        {"type": "fax", "value": "111-222-3333", "note": "office"}
    ]


def test_add_link():
    m = GenericModel()
    m.add_link("http://example.com/1")
    m.add_link("http://example.com/2", note="xyz")
    assert m.links == [
        {"url": "http://example.com/1", "note": ""},
        {"url": "http://example.com/2", "note": "xyz"},
    ]


def test_add_associated_link_match():
    m = GenericModel()
    m._add_associated_link(
        "_associated",
        "something",
        "http://example.com/1.txt",
        media_type="text/plain",
        on_duplicate="error",
    )
    m._add_associated_link(
        "_associated",
        "something",
        "http://example.com/1.pdf",
        media_type="application/pdf",
        on_duplicate="error",
    )
    # one 'document' added, multiple links for it
    assert len(m._associated) == 1
    assert len(m._associated[0]["links"]) == 2


def test_add_associated_link_on_duplicate_bad():
    m = GenericModel()

    with pytest.raises(ValueError):
        m._add_associated_link(
            "_associated",
            "something",
            "http://example.com",
            media_type="text/html",
            on_duplicate="idk",
        )


def test_add_associated_link_on_duplicate_error():
    m = GenericModel()
    m._add_associated_link(
        "_associated",
        "something",
        "http://example.com",
        media_type="text/html",
        on_duplicate="error",
    )

    with pytest.raises(ValueError):
        m._add_associated_link(
            "_associated",
            "something else",
            "http://example.com",
            media_type="text/html",
            on_duplicate="error",
        )


def test_add_associated_link_on_duplicate_ignore():
    m = GenericModel()
    m._add_associated_link(
        "_associated",
        "something",
        "http://example.com",
        media_type="text/html",
        on_duplicate="ignore",
    )
    m._add_associated_link(
        "_associated",
        "something else",
        "http://example.com",
        media_type="text/html",
        on_duplicate="ignore",
    )
    # one 'document' added, single link for it, keeps first name
    assert len(m._associated) == 1
    assert len(m._associated[0]["links"]) == 1
    assert m._associated[0]["note"] == "something"


def test_add_name():
    m = GenericModel()

    m.add_name("Thiston", note="What my friends call me")

    assert m.other_names == [{"name": "Thiston", "note": "What my friends call me"}]

    m.add_name(
        "Johnseph Q. Publico",
        note="Birth name",
        start_date="1920-01",
        end_date="1949-12-31",
    )

    assert m.other_names == [
        {"name": "Thiston", "note": "What my friends call me"},
        {
            "name": "Johnseph Q. Publico",
            "note": "Birth name",
            "start_date": "1920-01",
            "end_date": "1949-12-31",
        },
    ]


def test_add_identifier():
    g = GenericModel()

    with pytest.raises(TypeError):
        g.add_identifier("id10t", foo="bar")

    g.add_identifier("id10t")
    g.add_identifier("l0l", scheme="kruft")

    assert g.identifiers[-1]["scheme"] == "kruft"
    assert g.identifiers[0]["identifier"] == "id10t"



import os
import importlib
import json
import uuid
import logging
import datetime
from collections import defaultdict, OrderedDict

import jsonschema
from jsonschema import Draft3Validator, FormatChecker
import scrapelib

from .. import utils, settings
from ..exceptions import ScrapeError, ScrapeValueError


@FormatChecker.cls_checks("uri-blank")
def uri_blank(value):
    return value == "" or FormatChecker().conforms(value, "uri")


@FormatChecker.cls_checks("uri")
def check_uri(val):
    return val and val.startswith(("http://", "https://", "ftp://"))


def cleanup_list(obj, default):
    if not obj:
        obj = default
    elif isinstance(obj, str):
        obj = [obj]
    elif not isinstance(obj, list):
        obj = list(obj)
    return sorted(obj)


def clean_whitespace(obj):
    """ deep whitespace clean for ScrapeObj & dicts """
    if isinstance(obj, dict):
        items = obj.items()
        use_setattr = False
    elif isinstance(obj, object):
        items = obj.__dict__.items()
        use_setattr = True

    for k, v in items:
        if isinstance(v, str) and v:
            newv = v.strip()
        elif isinstance(v, list) and v:
            if not v:
                continue
            elif isinstance(v[0], str):
                newv = [i.strip() for i in v]
            elif isinstance(v[0], (dict, object)):
                newv = [clean_whitespace(i) for i in v]
            else:
                raise ValueError(f"Unhandled case, {k} is list of {type(v[0])}")
        else:
            continue

        if use_setattr:
            setattr(obj, k, newv)
        else:
            obj[k] = newv

    return obj


class Scraper(scrapelib.Scraper):
    """ Base class for all scrapers """

    def __init__(
        self, jurisdiction, datadir, *, strict_validation=True, fastmode=False
    ):
        super(Scraper, self).__init__()

        # set options
        self.jurisdiction = jurisdiction
        self.datadir = datadir

        # scrapelib setup
        self.timeout = settings.SCRAPELIB_TIMEOUT
        self.requests_per_minute = settings.SCRAPELIB_RPM
        self.retry_attempts = settings.SCRAPELIB_RETRY_ATTEMPTS
        self.retry_wait_seconds = settings.SCRAPELIB_RETRY_WAIT_SECONDS
        self.verify = settings.SCRAPELIB_VERIFY

        # caching
        if settings.CACHE_DIR:
            self.cache_storage = scrapelib.FileCache(settings.CACHE_DIR)

        if fastmode:
            self.requests_per_minute = 0
            self.cache_write_only = False

        # validation
        self.strict_validation = strict_validation

        # 'type' -> {set of names}
        self.output_names = defaultdict(set)

        # logging convenience methods
        self.logger = logging.getLogger("openstates")
        self.info = self.logger.info
        self.debug = self.logger.debug
        self.warning = self.logger.warning
        self.error = self.logger.error
        self.critical = self.logger.critical

        modname = os.environ.get("SCRAPE_OUTPUT_HANDLER")
        if modname is None:
            self.scrape_output_handler = None
        else:
            handler = importlib.import_module(modname)
            self.scrape_output_handler = handler.Handler(self)

    def save_object(self, obj):
        """
        Save object to disk as JSON.

        Generally shouldn't be called directly.
        """
        clean_whitespace(obj)
        obj.pre_save(self.jurisdiction.jurisdiction_id)

        filename = "{0}_{1}.json".format(obj._type, obj._id).replace("/", "-")

        self.info("save %s %s as %s", obj._type, obj, filename)
        self.debug(
            json.dumps(
                OrderedDict(sorted(obj.as_dict().items())),
                cls=utils.JSONEncoderPlus,
                indent=4,
                separators=(",", ": "),
            )
        )

        self.output_names[obj._type].add(filename)

        if self.scrape_output_handler is None:
            with open(os.path.join(self.datadir, filename), "w") as f:
                json.dump(obj.as_dict(), f, cls=utils.JSONEncoderPlus)
        else:
            self.scrape_output_handler.handle(obj)

        # validate after writing, allows for inspection on failure
        try:
            obj.validate()
        except ValueError as ve:
            if self.strict_validation:
                raise ve
            else:
                self.warning(ve)

        # after saving and validating, save subordinate objects
        for obj in obj._related:
            self.save_object(obj)

    def do_scrape(self, **kwargs):
        record = {"objects": defaultdict(int)}
        self.output_names = defaultdict(set)
        record["start"] = utils.utcnow()
        for obj in self.scrape(**kwargs) or []:
            if hasattr(obj, "__iter__"):
                for iterobj in obj:
                    self.save_object(iterobj)
            else:
                self.save_object(obj)
        record["end"] = utils.utcnow()
        record["skipped"] = getattr(self, "skipped", 0)
        if not self.output_names:
            raise ScrapeError(
                "no objects returned from {} scrape".format(self.__class__.__name__)
            )
        for _type, nameset in self.output_names.items():
            record["objects"][_type] += len(nameset)

        return record

    def latest_session(self):
        return self.jurisdiction.legislative_sessions[-1]["identifier"]

    def scrape(self, **kwargs):
        raise NotImplementedError(
            self.__class__.__name__ + " must provide a scrape() method"
        )


class BaseBillScraper(Scraper):
    skipped = 0

    class ContinueScraping(Exception):
        """ indicate that scraping should continue without saving an object """

        pass

    def scrape(self, legislative_session, **kwargs):
        self.legislative_session = legislative_session
        for bill_id, extras in self.get_bill_ids(**kwargs):
            try:
                yield self.get_bill(bill_id, **extras)
            except self.ContinueScraping as exc:
                self.warning("skipping %s: %r", bill_id, exc)
                self.skipped += 1
                continue


class BaseModel(object):
    """
    This is the base class for all the Open Civic objects. This contains
    common methods and abstractions for OCD objects.
    """

    # to be overridden by children. Something like "person" or "organization".
    # Used in :func:`validate`.
    _type = None
    _schema = None

    def __init__(self):
        super(BaseModel, self).__init__()
        self._id = str(uuid.uuid1())
        self._related = []
        self.extras = {}

    # validation

    def validate(self, schema=None):
        """
        Validate that we have a valid object.

        On error, this will raise a `ScrapeValueError`

        This also expects that the schemas assume that omitting required
        in the schema asserts the field is optional, not required. This is
        due to upstream schemas being in JSON Schema v3, and not validictory's
        modified syntax.
        ^ TODO: FIXME
        """
        if schema is None:
            schema = self._schema

        type_checker = Draft3Validator.TYPE_CHECKER.redefine(
            "datetime", lambda c, d: isinstance(d, (datetime.date, datetime.datetime))
        )
        type_checker = type_checker.redefine(
            "date",
            lambda c, d: (
                isinstance(d, datetime.date) and not isinstance(d, datetime.datetime)
            ),
        )

        ValidatorCls = jsonschema.validators.extend(
            Draft3Validator, type_checker=type_checker
        )
        validator = ValidatorCls(schema, format_checker=FormatChecker())

        errors = [str(error) for error in validator.iter_errors(self.as_dict())]
        if errors:
            raise ScrapeValueError(
                "validation of {} {} failed: {}".format(
                    self.__class__.__name__, self._id, "\n\t" + "\n\t".join(errors)
                )
            )

    def pre_save(self, jurisdiction_id):
        pass

    def as_dict(self):
        d = {}
        for attr in self._schema["properties"].keys():
            if hasattr(self, attr):
                d[attr] = getattr(self, attr)
        d["_id"] = self._id
        return d

    # operators

    def __setattr__(self, key, val):
        if key[0] != "_" and key not in self._schema["properties"].keys():
            raise ScrapeValueError(
                'property "{}" not in {} schema'.format(key, self._type)
            )
        super(BaseModel, self).__setattr__(key, val)


class SourceMixin(object):
    def __init__(self):
        super(SourceMixin, self).__init__()
        self.sources = []

    def add_source(self, url, *, note=""):
        """ Add a source URL from which data was collected """
        new = {"url": url, "note": note}
        self.sources.append(new)


class ContactDetailMixin(object):
    def __init__(self):
        super(ContactDetailMixin, self).__init__()
        self.contact_details = []

    def add_contact_detail(self, *, type, value, note=""):
        self.contact_details.append({"type": type, "value": value, "note": note})


class LinkMixin(object):
    def __init__(self):
        super(LinkMixin, self).__init__()
        self.links = []

    def add_link(self, url, *, note=""):
        self.links.append({"note": note, "url": url})


class IdentifierMixin(object):
    def __init__(self):
        super(IdentifierMixin, self).__init__()
        self.identifiers = []

    def add_identifier(self, identifier, *, scheme=""):
        self.identifiers.append({"identifier": identifier, "scheme": scheme})


class OtherNameMixin(object):
    def __init__(self):
        super(OtherNameMixin, self).__init__()
        self.other_names = []

    def add_name(self, name, *, start_date="", end_date="", note=""):
        other_name = {"name": name}
        if start_date:
            other_name["start_date"] = start_date
        if end_date:
            other_name["end_date"] = end_date
        if note:
            other_name["note"] = note
        self.other_names.append(other_name)


class AssociatedLinkMixin(object):
    def _add_associated_link(
        self,
        collection,
        note,
        url,
        *,
        media_type,
        on_duplicate="warn",
        date="",
        classification="",
    ):
        if on_duplicate not in ["error", "ignore", "warn"]:
            raise ScrapeValueError("on_duplicate must be 'warn', 'error' or 'ignore'")

        try:
            associated = getattr(self, collection)
        except AttributeError:
            associated = self[collection]

        ver = {
            "note": note,
            "links": [],
            "date": date,
            "classification": classification,
        }

        # keep a list of the links we've seen, we need to iterate over whole list on each add
        # unfortunately this means adds are O(n)
        seen_links = set()

        matches = 0
        for item in associated:
            for link in item["links"]:
                seen_links.add(link["url"])

            if all(
                ver.get(x) == item.get(x) for x in ["note", "date", "classification"]
            ):
                matches = matches + 1
                ver = item

        # it should be impossible to have multiple matches found unless someone is bypassing
        # _add_associated_link
        assert matches <= 1, "multiple matches found in _add_associated_link"

        if url in seen_links:
            if on_duplicate == "error":
                raise ScrapeValueError(
                    "Duplicate entry in '%s' - URL: '%s'" % (collection, url)
                )
            elif on_duplicate == "warn":
                # default behavior: same as ignore but logs an warning so people can fix
                logging.getLogger("openstates").warning(
                    f"Duplicate entry in '{collection}' - URL: {url}"
                )
                return None
            else:
                # This means we're in ignore mode. This situation right here
                # means we should *skip* adding this link silently and continue
                # on with our scrape. This should *ONLY* be used when there's
                # a site issue (Version 1 == Version 2 because of a bug) and
                # *NEVER* because "Current" happens to match "Version 3". Fix
                # that in the scraper, please.
                #  - PRT
                return None

        # OK. This is either new or old. Let's just go for it.
        ret = {"url": url, "media_type": media_type}

        ver["links"].append(ret)

        if matches == 0:
            # in the event we've got a new entry; let's just insert it into
            # the versions on this object. Otherwise it'll get thrown in
            # automagically.
            associated.append(ver)

        return ver



from django.db.models import Q
from ._types import _JsonDict
from .base import BaseImporter
from ..data.models import Organization


class OrganizationImporter(BaseImporter):
    _type = "organization"
    model_class = Organization

    def limit_spec(self, spec: _JsonDict) -> _JsonDict:
        if spec.get("classification") != "party":
            spec["jurisdiction_id"] = self.jurisdiction_id

        name = spec.pop("name", None)
        if name:
            return Q(**spec) & Q(name=name)
        return spec



from django.db import models
from django.contrib.postgres.fields import ArrayField, JSONField
from .base import (
    OCDBase,
    LinkBase,
    OCDIDField,
    RelatedBase,
    RelatedEntityBase,
    MimetypeLinkBase,
)
from .jurisdiction import Jurisdiction
from .bill import Bill
from .vote import VoteEvent
from ..common import (
    EVENT_MEDIA_CLASSIFICATION_CHOICES,
    EVENT_DOCUMENT_CLASSIFICATION_CHOICES,
)

EVENT_STATUS_CHOICES = (
    ("cancelled", "Cancelled"),
    ("tentative", "Tentative"),
    ("confirmed", "Confirmed"),
    ("passed", "Passed"),
)


class EventMediaBase(RelatedBase):
    note = models.CharField(max_length=300)
    date = models.CharField(max_length=25, blank=True)  # YYYY-MM-DD HH:MM:SS+HH:MM
    offset = models.PositiveIntegerField(null=True)

    class Meta:
        abstract = True


class EventLocation(RelatedBase):
    name = models.CharField(max_length=200)
    url = models.URLField(blank=True, max_length=2000)
    jurisdiction = models.ForeignKey(
        Jurisdiction, related_name="event_locations", on_delete=models.CASCADE
    )

    def __str__(self):
        return self.name

    class Meta:
        db_table = "opencivicdata_eventlocation"


class Event(OCDBase):
    id = OCDIDField(ocd_type="event")
    name = models.CharField(max_length=1000)
    jurisdiction = models.ForeignKey(
        Jurisdiction,
        related_name="events",
        # jurisdictions hard to delete
        on_delete=models.PROTECT,
    )
    description = models.TextField()
    classification = models.CharField(max_length=100)
    start_date = models.CharField(max_length=25)  # YYYY-MM-DD HH:MM:SS+HH:MM
    end_date = models.CharField(max_length=25, blank=True)  # YYYY-MM-DD HH:MM:SS+HH:MM
    all_day = models.BooleanField(default=False)
    status = models.CharField(max_length=20, choices=EVENT_STATUS_CHOICES)
    location = models.ForeignKey(EventLocation, null=True, on_delete=models.SET_NULL)
    dedupe_key = models.CharField(max_length=500, null=True)

    def __str__(self):
        return "{0} ({1})".format(self.name, self.start_date)

    class Meta:
        db_table = "opencivicdata_event"
        index_together = [["jurisdiction", "start_date", "name"]]


class EventMedia(EventMediaBase):
    event = models.ForeignKey(Event, related_name="media", on_delete=models.CASCADE)
    classification = models.CharField(
        max_length=50, choices=EVENT_MEDIA_CLASSIFICATION_CHOICES, blank=True
    )

    def __str__(self):
        return "%s for %s" % (self.note, self.event)

    class Meta:
        db_table = "opencivicdata_eventmedia"


class EventMediaLink(MimetypeLinkBase):
    media = models.ForeignKey(
        EventMedia, related_name="links", on_delete=models.CASCADE
    )

    def __str__(self):
        return "{0} for {1}".format(self.url, self.media.event)

    class Meta:
        db_table = "opencivicdata_eventmedialink"


class EventDocument(RelatedBase):
    event = models.ForeignKey(Event, related_name="documents", on_delete=models.CASCADE)
    note = models.CharField(max_length=300)
    date = models.CharField(max_length=25, blank=True)  # YYYY-MM-DD HH:MM:SS+HH:MM
    classification = models.CharField(
        max_length=50, choices=EVENT_DOCUMENT_CLASSIFICATION_CHOICES, blank=True
    )

    def __str__(self):
        tmpl = "{doc.note} for event {doc.event}"
        return tmpl.format(doc=self)

    class Meta:
        db_table = "opencivicdata_eventdocument"


class EventDocumentLink(MimetypeLinkBase):
    document = models.ForeignKey(
        EventDocument, related_name="links", on_delete=models.CASCADE
    )

    def __str__(self):
        return "{0} for {1}".format(self.url, self.document)

    class Meta:
        db_table = "opencivicdata_eventdocumentlink"


class EventLink(LinkBase):
    event = models.ForeignKey(Event, related_name="links", on_delete=models.CASCADE)

    class Meta:
        db_table = "opencivicdata_eventlink"


class EventSource(LinkBase):
    event = models.ForeignKey(Event, related_name="sources", on_delete=models.CASCADE)

    class Meta:
        db_table = "opencivicdata_eventsource"


class EventParticipant(RelatedEntityBase):
    event = models.ForeignKey(
        Event, related_name="participants", on_delete=models.CASCADE
    )
    note = models.TextField()

    def __str__(self):
        tmpl = "%s at %s"
        return tmpl % (self.name, self.event)

    class Meta:
        db_table = "opencivicdata_eventparticipant"


class EventAgendaItem(RelatedBase):
    description = models.TextField()
    classification = ArrayField(base_field=models.TextField(), blank=True, default=list)
    order = models.PositiveIntegerField()
    subjects = ArrayField(base_field=models.TextField(), blank=True, default=list)
    notes = ArrayField(base_field=models.TextField(), blank=True, default=list)
    event = models.ForeignKey(Event, related_name="agenda", on_delete=models.CASCADE)
    extras = JSONField(default=dict, blank=True)

    def __str__(self):
        return "Agenda item {0} for {1}".format(self.order, self.event).replace(
            "  ", " "
        )

    class Meta:
        db_table = "opencivicdata_eventagendaitem"


class EventRelatedEntity(RelatedEntityBase):
    agenda_item = models.ForeignKey(
        EventAgendaItem, related_name="related_entities", on_delete=models.CASCADE
    )
    # will just unresolve if needed
    bill = models.ForeignKey(Bill, null=True, on_delete=models.SET_NULL)
    vote_event = models.ForeignKey(VoteEvent, null=True, on_delete=models.SET_NULL)
    note = models.TextField()

    def __str__(self):
        return "{0} related to {1}".format(self.entity_name, self.agenda_item)

    @property
    def entity_name(self):
        if self.entity_type == "vote" and self.vote_event_id:
            return self.vote_event.identifier
        elif self.entity_type == "bill" and self.bill_id:
            return self.bill.identifier
        else:
            return super(EventRelatedEntity, self).entity_name

    @property
    def entity_id(self):
        if self.entity_type == "vote":
            return self.vote_event_id
        if self.entity_type == "bill":
            return self.bill_id
        return super(EventRelatedEntity, self).entity_id

    class Meta:
        db_table = "opencivicdata_eventrelatedentity"


class EventAgendaMedia(EventMediaBase):
    agenda_item = models.ForeignKey(
        EventAgendaItem, related_name="media", on_delete=models.CASCADE
    )
    classification = models.CharField(max_length=100, blank=True, default="")

    def __str__(self):
        return "{0} for {1}".format(self.note, self.agenda_item)

    class Meta:
        db_table = "opencivicdata_eventagendamedia"


class EventAgendaMediaLink(MimetypeLinkBase):
    media = models.ForeignKey(
        EventAgendaMedia, related_name="links", on_delete=models.CASCADE
    )

    def __str__(self):
        return "{0} for {1}".format(self.url, self.media)

    class Meta:
        db_table = "opencivicdata_eventagendamedialink"



from ..metadata import STATES_BY_ABBR
from ..utils.django import init_django
from django.db import transaction


def create_division(division_id, name):
    from ..data.models import Division

    return Division.objects.get_or_create(
        id=division_id,
        country="us",
        # TODO: allow changing name
        defaults=dict(name=name),
    )[0]


def create_chamber(juris, parent, chamber):
    from ..data.models import Organization, Post

    if chamber.chamber_type != "unicameral":
        post_parent, created = Organization.objects.get_or_create(
            id=chamber.organization_id,
            classification=chamber.chamber_type,
            parent_id=parent.id,
            jurisdiction_id=juris.id,
            name=chamber.name,
        )
    else:
        # parent is unicameral org
        post_parent = parent

    # create divisions and posts
    for district in chamber.districts:
        post_div = create_division(
            district.division_id, f"{juris.name} {chamber.name} {district.name}"
        )
        Post.objects.get_or_create(
            label=district.name,
            organization=post_parent,
            division=post_div,
            # TODO: allow changing role & max_memberships
            defaults=dict(
                role=district.title_override or chamber.title,
                maximum_memberships=district.num_seats,
            ),
        )


def create_full_jurisdiction(state):
    from ..data.models import Jurisdiction, Organization

    div = create_division(state.division_id, state.name)
    juris, created = Jurisdiction.objects.get_or_create(
        id=state.jurisdiction_id,
        name=state.name,
        division=div,
        classification="state" if state.abbr != "US" else "country",
        defaults=dict(url=state.url),
    )
    leg, created = Organization.objects.get_or_create(
        id=state.legislature_organization_id,
        classification="legislature",
        jurisdiction=juris,
        defaults=dict(name=state.legislature_name),
    )
    # create executive
    Organization.objects.get_or_create(
        id=state.executive_organization_id,
        classification="executive",
        jurisdiction=juris,
        defaults=dict(name=state.executive_name),
    )

    if state.unicameral:
        create_chamber(juris, leg, state.legislature)
    else:
        create_chamber(juris, leg, state.lower)
        create_chamber(juris, leg, state.upper)


def load_jurisdictions():
    for name, state in STATES_BY_ABBR.items():
        print("loading", name)
        create_full_jurisdiction(state)


def main():
    init_django()
    with transaction.atomic():
        load_jurisdictions()



# Generated by Django 3.0.5 on 2020-04-23 11:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("data", "0004_auto_20200422_0040")]

    operations = [
        migrations.AddField(
            model_name="person",
            name="current_role_division_id",
            field=models.CharField(default="", max_length=100),
        ),
        migrations.AddField(
            model_name="person",
            name="primary_party",
            field=models.CharField(
                default="",
                help_text="Primary party an individual is associated with.",
                max_length=100,
            ),
        ),
    ]



from ..models import State, Chamber, District

PR = State(
    name="Puerto Rico",
    abbr="PR",
    capital="San Juan",
    capital_tz="America/Puerto_Rico",
    fips="72",
    unicameral=False,
    legislature_name="Legislative Assembly of Puerto Rico",
    legislature_organization_id="ocd-organization/172e46ae-bd24-406a-a754-e243e5fab319",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/c0dbab0c-4fe6-464e-aa44-6bd2cf725c55",
    division_id="ocd-division/country:us/territory:pr",
    jurisdiction_id="ocd-jurisdiction/country:us/territory:pr/government",
    url="http://www.oslpr.org/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/f932d329-15b2-4385-9d7d-d4b8a7ed8495",
        num_seats=51,
        title="Representative",
        districts=[
            District("1", "lower", "ocd-division/country:us/territory:pr/sldl:1", 1),
            District("2", "lower", "ocd-division/country:us/territory:pr/sldl:2", 1),
            District("3", "lower", "ocd-division/country:us/territory:pr/sldl:3", 1),
            District("4", "lower", "ocd-division/country:us/territory:pr/sldl:4", 1),
            District("5", "lower", "ocd-division/country:us/territory:pr/sldl:5", 1),
            District("6", "lower", "ocd-division/country:us/territory:pr/sldl:6", 1),
            District("7", "lower", "ocd-division/country:us/territory:pr/sldl:7", 1),
            District("8", "lower", "ocd-division/country:us/territory:pr/sldl:8", 1),
            District("9", "lower", "ocd-division/country:us/territory:pr/sldl:9", 1),
            District("10", "lower", "ocd-division/country:us/territory:pr/sldl:10", 1),
            District("11", "lower", "ocd-division/country:us/territory:pr/sldl:11", 1),
            District("12", "lower", "ocd-division/country:us/territory:pr/sldl:12", 1),
            District("13", "lower", "ocd-division/country:us/territory:pr/sldl:13", 1),
            District("14", "lower", "ocd-division/country:us/territory:pr/sldl:14", 1),
            District("15", "lower", "ocd-division/country:us/territory:pr/sldl:15", 1),
            District("16", "lower", "ocd-division/country:us/territory:pr/sldl:16", 1),
            District("17", "lower", "ocd-division/country:us/territory:pr/sldl:17", 1),
            District("18", "lower", "ocd-division/country:us/territory:pr/sldl:18", 1),
            District("19", "lower", "ocd-division/country:us/territory:pr/sldl:19", 1),
            District("20", "lower", "ocd-division/country:us/territory:pr/sldl:20", 1),
            District("21", "lower", "ocd-division/country:us/territory:pr/sldl:21", 1),
            District("22", "lower", "ocd-division/country:us/territory:pr/sldl:22", 1),
            District("23", "lower", "ocd-division/country:us/territory:pr/sldl:23", 1),
            District("24", "lower", "ocd-division/country:us/territory:pr/sldl:24", 1),
            District("25", "lower", "ocd-division/country:us/territory:pr/sldl:25", 1),
            District("26", "lower", "ocd-division/country:us/territory:pr/sldl:26", 1),
            District("27", "lower", "ocd-division/country:us/territory:pr/sldl:27", 1),
            District("28", "lower", "ocd-division/country:us/territory:pr/sldl:28", 1),
            District("29", "lower", "ocd-division/country:us/territory:pr/sldl:29", 1),
            District("30", "lower", "ocd-division/country:us/territory:pr/sldl:30", 1),
            District("31", "lower", "ocd-division/country:us/territory:pr/sldl:31", 1),
            District("32", "lower", "ocd-division/country:us/territory:pr/sldl:32", 1),
            District("33", "lower", "ocd-division/country:us/territory:pr/sldl:33", 1),
            District("34", "lower", "ocd-division/country:us/territory:pr/sldl:34", 1),
            District("35", "lower", "ocd-division/country:us/territory:pr/sldl:35", 1),
            District("36", "lower", "ocd-division/country:us/territory:pr/sldl:36", 1),
            District("37", "lower", "ocd-division/country:us/territory:pr/sldl:37", 1),
            District("38", "lower", "ocd-division/country:us/territory:pr/sldl:38", 1),
            District("39", "lower", "ocd-division/country:us/territory:pr/sldl:39", 1),
            District("40", "lower", "ocd-division/country:us/territory:pr/sldl:40", 1),
            District(
                "At-Large",
                "lower",
                "ocd-division/country:us/territory:pr/sldl:at-large",
                11,
            ),
        ],
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/c187ea65-5cd6-42fb-b08b-18e487855526",
        num_seats=27,
        title="Senator",
        districts=[
            District("I", "upper", "ocd-division/country:us/territory:pr/sldu:1", 2),
            District("II", "upper", "ocd-division/country:us/territory:pr/sldu:2", 2),
            District("III", "upper", "ocd-division/country:us/territory:pr/sldu:3", 2),
            District("IV", "upper", "ocd-division/country:us/territory:pr/sldu:4", 2),
            District("V", "upper", "ocd-division/country:us/territory:pr/sldu:5", 2),
            District("VI", "upper", "ocd-division/country:us/territory:pr/sldu:6", 2),
            District("VII", "upper", "ocd-division/country:us/territory:pr/sldu:7", 2),
            District("VIII", "upper", "ocd-division/country:us/territory:pr/sldu:8", 2),
            District("At-Large", "upper", "ocd-division/country:us/territory:pr", 11),
        ],
    ),
)
